#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Mar 16 18:04:55 2020

@author: ljia
"""
import numpy as np
from gklearn.ged.env import AlgorithmState, NodeMap
from gklearn.ged.util import misc
from gklearn.utils import Timer
import time
from tqdm import tqdm
import sys
import networkx as nx
import multiprocessing
from multiprocessing import Pool
from functools import partial


class MedianGraphEstimator(object): # @todo: differ dummy_node from undifined node?
	
	def __init__(self, ged_env, constant_node_costs):
		"""Constructor.
		
		Parameters
		----------
		ged_env : gklearn.gedlib.gedlibpy.GEDEnv
			Initialized GED environment. The edit costs must be set by the user.
			
		constant_node_costs : Boolean
			Set to True if the node relabeling costs are constant.
		"""
		self._ged_env = ged_env
		self._init_method = 'BRANCH_FAST'
		self._init_options = ''
		self._descent_method = 'BRANCH_FAST'
		self._descent_options = ''
		self._refine_method = 'IPFP'
		self._refine_options = ''
		self._constant_node_costs = constant_node_costs
		self._labeled_nodes = (ged_env.get_num_node_labels() > 1)
		self._node_del_cost = ged_env.get_node_del_cost(ged_env.get_node_label(1))
		self._node_ins_cost = ged_env.get_node_ins_cost(ged_env.get_node_label(1))
		self._labeled_edges = (ged_env.get_num_edge_labels() > 1)
		self._edge_del_cost = ged_env.get_edge_del_cost(ged_env.get_edge_label(1))
		self._edge_ins_cost = ged_env.get_edge_ins_cost(ged_env.get_edge_label(1))
		self._init_type = 'RANDOM'
		self._num_random_inits = 10
		self._desired_num_random_inits = 10
		self._use_real_randomness = True
		self._seed = 0
		self._parallel = True
		self._update_order = True
		self._sort_graphs = True # sort graphs by size when computing GEDs.
		self._refine = True
		self._time_limit_in_sec = 0
		self._epsilon = 0.0001
		self._max_itrs = 100
		self._max_itrs_without_update = 3
		self._num_inits_increase_order = 10
		self._init_type_increase_order = 'K-MEANS++'
		self._max_itrs_increase_order = 10
		self._print_to_stdout = 2
		self._median_id = np.inf # @todo: check
		self._node_maps_from_median = {}
		self._sum_of_distances = 0
		self._best_init_sum_of_distances = np.inf
		self._converged_sum_of_distances = np.inf
		self._runtime = None
		self._runtime_initialized = None
		self._runtime_converged = None
		self._itrs = [] # @todo: check: {} ?
		self._num_decrease_order = 0
		self._num_increase_order = 0
		self._num_converged_descents = 0
		self._state = AlgorithmState.TERMINATED
		self._label_names = {}
		
		if ged_env is None:
			raise Exception('The GED environment pointer passed to the constructor of MedianGraphEstimator is null.')
		elif not ged_env.is_initialized():
			raise Exception('The GED environment is uninitialized. Call gedlibpy.GEDEnv.init() before passing it to the constructor of MedianGraphEstimator.')
	
	
	def set_options(self, options):
		"""Sets the options of the estimator.

		Parameters
		----------
		options : string
			String that specifies with which options to run the estimator.
		"""
		self._set_default_options()
		options_map = misc.options_string_to_options_map(options)
		for opt_name, opt_val in options_map.items():
			if opt_name == 'init-type':
				self._init_type = opt_val
				if opt_val != 'MEDOID' and opt_val != 'RANDOM' and opt_val != 'MIN' and opt_val != 'MAX' and opt_val != 'MEAN':
					raise Exception('Invalid argument ' + opt_val + ' for option init-type. Usage: options = "[--init-type RANDOM|MEDOID|EMPTY|MIN|MAX|MEAN] [...]"')
			elif opt_name == 'random-inits':
				try:
					self._num_random_inits = int(opt_val)
					self._desired_num_random_inits = self._num_random_inits
				except:
					raise Exception('Invalid argument "' + opt_val + '" for option random-inits. Usage: options = "[--random-inits <convertible to int greater 0>]"')

				if self._num_random_inits <= 0:
					raise Exception('Invalid argument "' + opt_val + '" for option random-inits. Usage: options = "[--random-inits <convertible to int greater 0>]"')
	
			elif opt_name == 'randomness':
				if opt_val == 'PSEUDO':
					self._use_real_randomness = False
	
				elif opt_val == 'REAL':
					self._use_real_randomness = True
	
				else:
					raise Exception('Invalid argument "' + opt_val  + '" for option randomness. Usage: options = "[--randomness REAL|PSEUDO] [...]"')
	
			elif opt_name == 'stdout':
				if opt_val == '0':
					self._print_to_stdout = 0
	
				elif opt_val == '1':
					self._print_to_stdout = 1
	
				elif opt_val == '2':
					self._print_to_stdout = 2
	
				else:
					raise Exception('Invalid argument "' + opt_val  + '" for option stdout. Usage: options = "[--stdout 0|1|2] [...]"')

			elif opt_name == 'parallel':
				if opt_val == 'TRUE':
					self._parallel = True
	
				elif opt_val == 'FALSE':
					self._parallel = False
	
				else:
					raise Exception('Invalid argument "' + opt_val  + '" for option parallel. Usage: options = "[--parallel TRUE|FALSE] [...]"')
	
			elif opt_name == 'update-order':
				if opt_val == 'TRUE':
					self._update_order = True
	
				elif opt_val == 'FALSE':
					self._update_order = False
	
				else:
					raise Exception('Invalid argument "' + opt_val  + '" for option update-order. Usage: options = "[--update-order TRUE|FALSE] [...]"')
					
			elif opt_name == 'sort-graphs':
				if opt_val == 'TRUE':
					self._sort_graphs = True
	
				elif opt_val == 'FALSE':
					self._sort_graphs = False
	
				else:
					raise Exception('Invalid argument "' + opt_val  + '" for option sort-graphs. Usage: options = "[--sort-graphs TRUE|FALSE] [...]"')
					
			elif opt_name == 'refine':
				if opt_val == 'TRUE':
					self._refine = True
	
				elif opt_val == 'FALSE':
					self._refine = False
	
				else:
					raise Exception('Invalid argument "' + opt_val  + '" for option refine. Usage: options = "[--refine TRUE|FALSE] [...]"')
	
			elif opt_name == 'time-limit':
				try:
					self._time_limit_in_sec = float(opt_val)
	
				except:
					raise Exception('Invalid argument "' + opt_val + '" for option time-limit.  Usage: options = "[--time-limit <convertible to double>] [...]')
	
			elif opt_name == 'max-itrs':
				try:
					self._max_itrs = int(opt_val)
	
				except:
					raise Exception('Invalid argument "' + opt_val + '" for option max-itrs. Usage: options = "[--max-itrs <convertible to int>] [...]')
	
			elif opt_name == 'max-itrs-without-update':
				try:
					self._max_itrs_without_update = int(opt_val)
	
				except:
					raise Exception('Invalid argument "' + opt_val + '" for option max-itrs-without-update. Usage: options = "[--max-itrs-without-update <convertible to int>] [...]')
	
			elif opt_name == 'seed':
				try:
					self._seed = int(opt_val)
	
				except:
					raise Exception('Invalid argument "' + opt_val + '" for option seed. Usage: options = "[--seed <convertible to int greater equal 0>] [...]')
	
			elif opt_name == 'epsilon':
				try:
					self._epsilon = float(opt_val)
	
				except:
					raise Exception('Invalid argument "' + opt_val + '" for option epsilon. Usage: options = "[--epsilon <convertible to double greater 0>] [...]')
	
				if self._epsilon <= 0:
					raise Exception('Invalid argument "' + opt_val + '" for option epsilon. Usage: options = "[--epsilon <convertible to double greater 0>] [...]')
	
			elif opt_name == 'inits-increase-order':
				try:
					self._num_inits_increase_order = int(opt_val)
	
				except:
					raise Exception('Invalid argument "' + opt_val + '" for option inits-increase-order. Usage: options = "[--inits-increase-order <convertible to int greater 0>]"')
	
				if self._num_inits_increase_order <= 0:
					raise Exception('Invalid argument "' + opt_val + '" for option inits-increase-order. Usage: options = "[--inits-increase-order <convertible to int greater 0>]"')

			elif opt_name == 'init-type-increase-order':
				self._init_type_increase_order = opt_val
				if opt_val != 'CLUSTERS' and opt_val != 'K-MEANS++':
					raise Exception('Invalid argument ' + opt_val + ' for option init-type-increase-order. Usage: options = "[--init-type-increase-order CLUSTERS|K-MEANS++] [...]"')
	
			elif opt_name == 'max-itrs-increase-order':
				try:
					self._max_itrs_increase_order = int(opt_val)
	
				except:
					raise Exception('Invalid argument "' + opt_val + '" for option max-itrs-increase-order. Usage: options = "[--max-itrs-increase-order <convertible to int>] [...]')

			else:
				valid_options = '[--init-type <arg>] [--random-inits <arg>] [--randomness <arg>] [--seed <arg>] [--stdout <arg>] '
				valid_options += '[--time-limit <arg>] [--max-itrs <arg>] [--epsilon <arg>] '
				valid_options += '[--inits-increase-order <arg>] [--init-type-increase-order <arg>] [--max-itrs-increase-order <arg>]'
				raise Exception('Invalid option "' + opt_name + '". Usage: options = "' + valid_options + '"')
 
		
	def set_init_method(self, init_method, init_options=''):
		"""Selects method to be used for computing the initial medoid graph.
		
		Parameters
		----------
		init_method : string
			The selected method. Default: ged::Options::GEDMethod::BRANCH_UNIFORM.
		
		init_options : string
			The options for the selected method. Default: "".
		
		Notes
		-----
		Has no effect unless "--init-type MEDOID" is passed to set_options().
		"""
		self._init_method = init_method;
		self._init_options = init_options;
	
	
	def set_descent_method(self, descent_method, descent_options=''):
		"""Selects method to be used for block gradient descent..
		
		Parameters
		----------
		descent_method : string
			The selected method. Default: ged::Options::GEDMethod::BRANCH_FAST.
		
		descent_options : string
			The options for the selected method. Default: "".
		
		Notes
		-----
		Has no effect unless "--init-type MEDOID" is passed to set_options().
		"""
		self._descent_method = descent_method;
		self._descent_options = descent_options;

	
	def set_refine_method(self, refine_method, refine_options):
		"""Selects method to be used for improving the sum of distances and the node maps for the converged median.
		
		Parameters
		----------
		refine_method : string
			The selected method. Default: "IPFP".
			
		refine_options : string 
			The options for the selected method. Default: "".
					
		Notes
		-----
		Has no effect if "--refine FALSE" is passed to set_options().
		"""
		self._refine_method = refine_method
		self._refine_options = refine_options

	
	def run(self, graph_ids, set_median_id, gen_median_id):
		"""Computes a generalized median graph.
		
		Parameters
		----------
		graph_ids : list[integer]
			The IDs of the graphs for which the median should be computed. Must have been added to the environment passed to the constructor.
		
		set_median_id : integer
			The ID of the computed set-median. A dummy graph with this ID must have been added to the environment passed to the constructor. Upon termination, the computed median can be obtained via gklearn.gedlib.gedlibpy.GEDEnv.get_graph().


		gen_median_id : integer
			The ID of the computed generalized median. Upon termination, the computed median can be obtained via gklearn.gedlib.gedlibpy.GEDEnv.get_graph().
		"""
		# Sanity checks.
		if len(graph_ids) == 0:
			raise Exception('Empty vector of graph IDs, unable to compute median.')
		all_graphs_empty = True
		for graph_id in graph_ids:
			if self._ged_env.get_graph_num_nodes(graph_id) > 0:
				all_graphs_empty = False
				break
		if all_graphs_empty:
			raise Exception('All graphs in the collection are empty.')
			
		# Start timer and record start time.
		start = time.time()
		timer = Timer(self._time_limit_in_sec)
		self._median_id = gen_median_id
		self._state = AlgorithmState.TERMINATED
		
		# Get NetworkX graph representations of the input graphs.
		graphs = {}
		for graph_id in graph_ids:
			# @todo: get_nx_graph() function may need to be modified according to the coming code.
			graphs[graph_id] = self._ged_env.get_nx_graph(graph_id, True, True, False)
#		print(self._ged_env.get_graph_internal_id(0))
#		print(graphs[0].graph)
#		print(graphs[0].nodes(data=True))
#		print(graphs[0].edges(data=True))
#		print(nx.adjacency_matrix(graphs[0]))
			
		# Construct initial medians.
		medians = []
		self._construct_initial_medians(graph_ids, timer, medians)
		end_init = time.time()
		self._runtime_initialized = end_init - start
		print(medians[0].graph)
		print(medians[0].nodes(data=True))
		print(medians[0].edges(data=True))
		print(nx.adjacency_matrix(medians[0]))
		
		# Reset information about iterations and number of times the median decreases and increases.
		self._itrs = [0] * len(medians)
		self._num_decrease_order = 0
		self._num_increase_order = 0
		self._num_converged_descents = 0
		
		# Initialize the best median.
		best_sum_of_distances = np.inf
		self._best_init_sum_of_distances = np.inf
		node_maps_from_best_median = {}
		
		# Run block gradient descent from all initial medians.
		self._ged_env.set_method(self._descent_method, self._descent_options)
		for median_pos in range(0, len(medians)):
			
			# Terminate if the timer has expired and at least one SOD has been computed.
			if timer.expired() and median_pos > 0:
				break
			
			# Print information about current iteration.
			if self._print_to_stdout == 2:
				print('\n===========================================================')
				print('Block gradient descent for initial median', str(median_pos + 1), 'of', str(len(medians)), '.')
				print('-----------------------------------------------------------')
				
			# Get reference to the median.
			median = medians[median_pos]
			
			# Load initial median into the environment.
			self._ged_env.load_nx_graph(median, gen_median_id)
			self._ged_env.init(self._ged_env.get_init_type())
			
			# Compute node maps and sum of distances for initial median.
			xxx = self._node_maps_from_median
			self._compute_init_node_maps(graph_ids, gen_median_id)
			yyy = self._node_maps_from_median
			
			self._best_init_sum_of_distances = min(self._best_init_sum_of_distances, self._sum_of_distances)
			self._ged_env.load_nx_graph(median, set_median_id)
			print(self._best_init_sum_of_distances)
				
			# Run block gradient descent from initial median.
			converged = False
			itrs_without_update = 0
			while not self._termination_criterion_met(converged, timer, self._itrs[median_pos], itrs_without_update):
				
				# Print information about current iteration.
				if self._print_to_stdout == 2:
					print('\n===========================================================')
					print('Iteration', str(self._itrs[median_pos] + 1), 'for initial median', str(median_pos + 1), 'of', str(len(medians)), '.')
					print('-----------------------------------------------------------')
					
				# Initialize flags that tell us what happened in the iteration.
				median_modified = False
				node_maps_modified = False
				decreased_order = False
				increased_order = False
				
				# Update the median.
				median_modified = self._update_median(graphs, median)
				if self._update_order:
					if not median_modified or self._itrs[median_pos] == 0:
						decreased_order = self._decrease_order(graphs, median)
						if not decreased_order or self._itrs[median_pos] == 0:
							increased_order = self._increase_order(graphs, median)
						
				# Update the number of iterations without update of the median.
				if median_modified or decreased_order or increased_order:
					itrs_without_update = 0
				else:
					itrs_without_update += 1
					
				# Print information about current iteration.
				if self._print_to_stdout == 2:
					print('Loading median to environment: ... ', end='')
					
				# Load the median into the environment.
				# @todo: should this function use the original node label?
				self._ged_env.load_nx_graph(median, gen_median_id)
				self._ged_env.init(self._ged_env.get_init_type())
					
				# Print information about current iteration.
				if self._print_to_stdout == 2:
					print('done.')					
					
				# Print information about current iteration.
				if self._print_to_stdout == 2:
					print('Updating induced costs: ... ', end='')

				# Compute induced costs of the old node maps w.r.t. the updated median.
				for graph_id in graph_ids:
# 					print(self._node_maps_from_median[graph_id].induced_cost())
# 					xxx = self._node_maps_from_median[graph_id]					   
					self._ged_env.compute_induced_cost(gen_median_id, graph_id, self._node_maps_from_median[graph_id])
# 					print('---------------------------------------')
# 					print(self._node_maps_from_median[graph_id].induced_cost())
					# @todo:!!!!!!!!!!!!!!!!!!!!!!!!!!!!This value is a slight different from the c++ program, which might be a bug! Use it very carefully!
					
				# Print information about current iteration.
				if self._print_to_stdout == 2:
					print('done.')					
					
				# Update the node maps.
				node_maps_modified = self._update_node_maps()

				# Update the order of the median if no improvement can be found with the current order.
				
				# Update the sum of distances.
				old_sum_of_distances = self._sum_of_distances
				self._sum_of_distances = 0
				for graph_id, node_map in self._node_maps_from_median.items():
					self._sum_of_distances += node_map.induced_cost()
# 					print(self._sum_of_distances)
					
				# Print information about current iteration.
				if self._print_to_stdout == 2:
					print('Old local SOD: ', old_sum_of_distances)
					print('New local SOD: ', self._sum_of_distances)
					print('Best converged SOD: ', best_sum_of_distances)
					print('Modified median: ', median_modified)
					print('Modified node maps: ', node_maps_modified)
					print('Decreased order: ', decreased_order)
					print('Increased order: ', increased_order)
					print('===========================================================\n')
					
				converged = not (median_modified or node_maps_modified or decreased_order or increased_order)
				
				self._itrs[median_pos] += 1
				
			# Update the best median.
			if self._sum_of_distances < best_sum_of_distances:
				best_sum_of_distances = self._sum_of_distances
				node_maps_from_best_median = self._node_maps_from_median.copy() # @todo: this is a shallow copy, not sure if it is enough.
				best_median = median
				
			# Update the number of converged descents.
			if converged:
				self._num_converged_descents += 1
				
		# Store the best encountered median.
		self._sum_of_distances = best_sum_of_distances
		self._node_maps_from_median = node_maps_from_best_median
		self._ged_env.load_nx_graph(best_median, gen_median_id)
		self._ged_env.init(self._ged_env.get_init_type())
		end_descent = time.time()
		self._runtime_converged = end_descent - start
		
		# Refine the sum of distances and the node maps for the converged median.
		self._converged_sum_of_distances = self._sum_of_distances
		if self._refine:
			self._improve_sum_of_distances(timer)
		
		# Record end time, set runtime and reset the number of initial medians.
		end = time.time()
		self._runtime = end - start
		self._num_random_inits = self._desired_num_random_inits
		
		# Print global information.
		if self._print_to_stdout != 0:
			print('\n===========================================================')
			print('Finished computation of generalized median graph.')
			print('-----------------------------------------------------------')
			print('Best SOD after initialization: ', self._best_init_sum_of_distances)
			print('Converged SOD: ', self._converged_sum_of_distances)
			if self._refine:
				print('Refined SOD: ', self._sum_of_distances)
			print('Overall runtime: ', self._runtime)
			print('Runtime of initialization: ', self._runtime_initialized)
			print('Runtime of block gradient descent: ', self._runtime_converged - self._runtime_initialized)
			if self._refine:
				print('Runtime of refinement: ', self._runtime - self._runtime_converged)
			print('Number of initial medians: ', len(medians))
			total_itr = 0
			num_started_descents = 0
			for itr in self._itrs:
				total_itr += itr
				if itr > 0:
					num_started_descents += 1
			print('Size of graph collection: ', len(graph_ids))
			print('Number of started descents: ', num_started_descents)
			print('Number of converged descents: ', self._num_converged_descents)
			print('Overall number of iterations: ', total_itr)
			print('Overall number of times the order decreased: ', self._num_decrease_order)
			print('Overall number of times the order increased: ', self._num_increase_order)
			print('===========================================================\n')
			
			
	def _improve_sum_of_distances(self, timer): # @todo: go through and test
		# Use method selected for refinement phase.
		self._ged_env.set_method(self._refine_method, self._refine_options)
		
		# Print information about current iteration.
		if self._print_to_stdout == 2:
			progress = tqdm(desc='Improving node maps', total=len(self._node_maps_from_median), file=sys.stdout)
			print('\n===========================================================')
			print('Improving node maps and SOD for converged median.')
			print('-----------------------------------------------------------')
			progress.update(1)
			
		# Improving the node maps.
		nb_nodes_median = self._ged_env.get_graph_num_nodes(self._gen_median_id)
		for graph_id, node_map in self._node_maps_from_median.items():
			if time.expired():
				if self._state == AlgorithmState.TERMINATED:
					self._state = AlgorithmState.CONVERGED
				break

			nb_nodes_g = self._ged_env.get_graph_num_nodes(graph_id)
			if nb_nodes_median <= nb_nodes_g or not self._sort_graphs:			
				self._ged_env.run_method(self._gen_median_id, graph_id)
				if self._ged_env.get_upper_bound(self._gen_median_id, graph_id) < node_map.induced_cost():
					self._node_maps_from_median[graph_id] = self._ged_env.get_node_map(self._gen_median_id, graph_id)
			else:
				self._ged_env.run_method(graph_id, self._gen_median_id)
				if self._ged_env.get_upper_bound(graph_id, self._gen_median_id) < node_map.induced_cost():
						node_map_tmp = self._ged_env.get_node_map(graph_id, self._gen_median_id)
						node_map_tmp.forward_map, node_map_tmp.backward_map = node_map_tmp.backward_map, node_map_tmp.forward_map
						self._node_maps_from_median[graph_id] = node_map_tmp	
			
			self._sum_of_distances += self._node_maps_from_median[graph_id].induced_cost()				

			# Print information.
			if self._print_to_stdout == 2:
				progress.update(1)

		self._sum_of_distances = 0.0
		for key, val in self._node_maps_from_median.items():
			self._sum_of_distances += val.induced_cost()
			
		# Print information.
		if self._print_to_stdout == 2:
			print('===========================================================\n')
			
	
	def _median_available(self):
		return self._median_id != np.inf
	
	
	def get_state(self):
		if not self._median_available():
			raise Exception('No median has been computed. Call run() before calling get_state().')
		return self._state
		
			
	def get_sum_of_distances(self, state=''):
		"""Returns the sum of distances.
		
		Parameters
		----------
		state : string
			The state of the estimator. Can be 'initialized' or 'converged'. Default: ""
			
		Returns
		-------
		float
			The sum of distances (SOD) of the median when the estimator was in the state `state` during the last call to run(). If `state` is not given, the converged SOD (without refinement) or refined SOD (with refinement) is returned.
		"""
		if not self._median_available():
			raise Exception('No median has been computed. Call run() before calling get_sum_of_distances().')
		if state == 'initialized':
			return self._best_init_sum_of_distances
		if state == 'converged':
			return self._converged_sum_of_distances
		return self._sum_of_distances


	def get_runtime(self, state):
		if not self._median_available():
			raise Exception('No median has been computed. Call run() before calling get_runtime().')
		if state == AlgorithmState.INITIALIZED:
			return self._runtime_initialized
		if state == AlgorithmState.CONVERGED:
			return self._runtime_converged
		return self._runtime
	

	def get_num_itrs(self):
		if not self._median_available():
			raise Exception('No median has been computed. Call run() before calling get_num_itrs().')
		return self._itrs


	def get_num_times_order_decreased(self):
		if not self._median_available():
			raise Exception('No median has been computed. Call run() before calling get_num_times_order_decreased().')
		return self._num_decrease_order		
	
	
	def get_num_times_order_increased(self):
		if not self._median_available():
			raise Exception('No median has been computed. Call run() before calling get_num_times_order_increased().')
		return self._num_increase_order		
	
	
	def get_num_converged_descents(self):
		if not self._median_available():
			raise Exception('No median has been computed. Call run() before calling get_num_converged_descents().')
		return self._num_converged_descents
	
	
	def get_ged_env(self):
		return self._ged_env
	
	
	def _set_default_options(self):
		self._init_type = 'RANDOM'
		self._num_random_inits = 10
		self._desired_num_random_inits = 10
		self._use_real_randomness = True
		self._seed = 0
		self._parallel = True
		self._update_order = True
		self._sort_graphs = True
		self._refine = True
		self._time_limit_in_sec = 0
		self._epsilon = 0.0001
		self._max_itrs = 100
		self._max_itrs_without_update = 3
		self._num_inits_increase_order = 10
		self._init_type_increase_order = 'K-MEANS++'
		self._max_itrs_increase_order = 10
		self._print_to_stdout = 2
		self._label_names = {}
		
		
	def _construct_initial_medians(self, graph_ids, timer, initial_medians):
		# Print information about current iteration.
		if self._print_to_stdout == 2:
			print('\n===========================================================')
			print('Constructing initial median(s).')
			print('-----------------------------------------------------------')
			
		# Compute or sample the initial median(s).
		initial_medians.clear()
		if self._init_type == 'MEDOID':
			self._compute_medoid(graph_ids, timer, initial_medians)
		elif self._init_type == 'MAX':
			pass # @todo
#			compute_max_order_graph_(graph_ids, initial_medians)
		elif self._init_type == 'MIN':
			pass # @todo
#			compute_min_order_graph_(graph_ids, initial_medians)
		elif self._init_type == 'MEAN':
			pass # @todo
#			compute_mean_order_graph_(graph_ids, initial_medians)
		else:
			pass # @todo
#			sample_initial_medians_(graph_ids, initial_medians)

		# Print information about current iteration.
		if self._print_to_stdout == 2:
			print('===========================================================')
			
			
	def _compute_medoid(self, graph_ids, timer, initial_medians):
		# Use method selected for initialization phase.
		self._ged_env.set_method(self._init_method, self._init_options)
					
		# Compute the medoid.
		if self._parallel:
			# @todo: notice when parallel self._ged_env is not modified.
			sum_of_distances_list = [np.inf] * len(graph_ids)
			len_itr = len(graph_ids)
			itr = zip(graph_ids, range(0, len(graph_ids)))
			n_jobs = multiprocessing.cpu_count()
			if len_itr < 100 * n_jobs:
				chunksize = int(len_itr / n_jobs) + 1
			else:
				chunksize = 100
			def init_worker(ged_env_toshare):
				global G_ged_env
				G_ged_env = ged_env_toshare
			do_fun = partial(_compute_medoid_parallel, graph_ids, self._sort_graphs)
			pool = Pool(processes=n_jobs, initializer=init_worker, initargs=(self._ged_env,))
			if self._print_to_stdout == 2:
				iterator = tqdm(pool.imap_unordered(do_fun, itr, chunksize),
							desc='Computing medoid', file=sys.stdout)
			else:
				iterator = pool.imap_unordered(do_fun, itr, chunksize)
			for i, dis in iterator:
				sum_of_distances_list[i] = dis
			pool.close()
			pool.join()
			
			medoid_id = np.argmin(sum_of_distances_list)
			best_sum_of_distances = sum_of_distances_list[medoid_id]
			
			initial_medians.append(self._ged_env.get_nx_graph(medoid_id, True, True, False)) # @todo

		else:
			# Print information about current iteration.
			self.ged_matrix_set_median_tmp = np.ones((len(graph_ids), len(graph_ids))) * np.inf
			if self._print_to_stdout == 2:
				progress = tqdm(desc='Computing medoid', total=len(graph_ids), file=sys.stdout)
		
			medoid_id = graph_ids[0]
			best_sum_of_distances = np.inf
			for g_id in graph_ids:
				if timer.expired():
					self._state = AlgorithmState.CALLED
					break
				nb_nodes_g = self._ged_env.get_graph_num_nodes(g_id)
				sum_of_distances = 0
				for h_id in graph_ids: # @todo: can this be faster?				
					nb_nodes_h = self._ged_env.get_graph_num_nodes(h_id)
					if nb_nodes_g <= nb_nodes_h or not self._sort_graphs:
						self._ged_env.run_method(g_id, h_id)
						sum_of_distances += self._ged_env.get_upper_bound(g_id, h_id)
						self.ged_matrix_set_median_tmp[g_id, h_id] = self._ged_env.get_upper_bound(g_id, h_id)
					else:
						# @todo: is this correct?
						self._ged_env.run_method(h_id, g_id)
						sum_of_distances += self._ged_env.get_upper_bound(h_id, g_id)
						self.ged_matrix_set_median_tmp[g_id, h_id] = self._ged_env.get_upper_bound(h_id, g_id)
					print(sum_of_distances)
				if sum_of_distances < best_sum_of_distances:
					best_sum_of_distances = sum_of_distances
					medoid_id = g_id
					
				# Print information about current iteration.
				if self._print_to_stdout == 2:
					progress.update(1)
					
			initial_medians.append(self._ged_env.get_nx_graph(medoid_id, True, True, False)) # @todo
			
			# Print information about current iteration.
			if self._print_to_stdout == 2:
				print('\n')
			
			
	def _compute_init_node_maps(self, graph_ids, gen_median_id):
		# Compute node maps and sum of distances for initial median.
		if self._parallel:
			# @todo: notice when parallel self._ged_env is not modified.
			self._sum_of_distances = 0
			self._node_maps_from_median.clear()
			sum_of_distances_list = [0] * len(graph_ids)
			
			len_itr = len(graph_ids)
			itr = graph_ids
			n_jobs = multiprocessing.cpu_count()
			if len_itr < 100 * n_jobs:
				chunksize = int(len_itr / n_jobs) + 1
			else:
				chunksize = 100
			def init_worker(ged_env_toshare):
				global G_ged_env
				G_ged_env = ged_env_toshare
			nb_nodes_median = self._ged_env.get_graph_num_nodes(gen_median_id)
			do_fun = partial(_compute_init_node_maps_parallel, gen_median_id, self._sort_graphs, nb_nodes_median)
			pool = Pool(processes=n_jobs, initializer=init_worker, initargs=(self._ged_env,))
			if self._print_to_stdout == 2:
				iterator = tqdm(pool.imap_unordered(do_fun, itr, chunksize),
							desc='Computing initial node maps', file=sys.stdout)
			else:
				iterator = pool.imap_unordered(do_fun, itr, chunksize)
			for g_id, sod, node_maps in iterator:
				sum_of_distances_list[g_id] = sod
				self._node_maps_from_median[g_id] = node_maps
			pool.close()
			pool.join()
			
			self._sum_of_distances = np.sum(sum_of_distances_list)
# 			xxx = self._node_maps_from_median
			
		else:
			# Print information about current iteration.
			if self._print_to_stdout == 2:
				progress = tqdm(desc='Computing initial node maps', total=len(graph_ids), file=sys.stdout)
				
			self._sum_of_distances = 0
			self._node_maps_from_median.clear()
			nb_nodes_median = self._ged_env.get_graph_num_nodes(gen_median_id)
			for graph_id in graph_ids:
				nb_nodes_g = self._ged_env.get_graph_num_nodes(graph_id)
				if nb_nodes_median <= nb_nodes_g or not self._sort_graphs:
					self._ged_env.run_method(gen_median_id, graph_id)
					self._node_maps_from_median[graph_id] = self._ged_env.get_node_map(gen_median_id, graph_id)
				else:
					self._ged_env.run_method(graph_id, gen_median_id)
					node_map_tmp = self._ged_env.get_node_map(graph_id, gen_median_id)
					node_map_tmp.forward_map, node_map_tmp.backward_map = node_map_tmp.backward_map, node_map_tmp.forward_map
					self._node_maps_from_median[graph_id] = node_map_tmp
	# 				print(self._node_maps_from_median[graph_id])
				self._sum_of_distances += self._node_maps_from_median[graph_id].induced_cost()
	# 				print(self._sum_of_distances)
				# Print information about current iteration.
				if self._print_to_stdout == 2:
					progress.update(1)
			
			# Print information about current iteration.
			if self._print_to_stdout == 2:
				print('\n')

		
	def _termination_criterion_met(self, converged, timer, itr, itrs_without_update):
		if timer.expired() or (itr >= self._max_itrs if self._max_itrs >= 0 else False):
			if self._state == AlgorithmState.TERMINATED:
				self._state = AlgorithmState.INITIALIZED
			return True
		return converged or (itrs_without_update > self._max_itrs_without_update if self._max_itrs_without_update >= 0 else False)
	
	
	def _update_median(self, graphs, median):
		# Print information about current iteration.
		if self._print_to_stdout == 2:
			print('Updating median: ', end='')
			
		# Store copy of the old median.
		old_median = median.copy() # @todo: this is just a shallow copy.
		
		# Update the node labels.
		if self._labeled_nodes:
			self._update_node_labels(graphs, median)
			
		# Update the edges and their labels.
		self._update_edges(graphs, median)
		
		# Print information about current iteration.
		if self._print_to_stdout == 2:
			print('done.')
			
		return not self._are_graphs_equal(median, old_median)
		
		
	def _update_node_labels(self, graphs, median):
# 		print('----------------------------')
		
		# Print information about current iteration.
		if self._print_to_stdout == 2:
			print('nodes ... ', end='')
			
		# Iterate through all nodes of the median.
		for i in range(0, nx.number_of_nodes(median)):
# 			print('i: ', i)
			# Collect the labels of the substituted nodes.
			node_labels = []
			for graph_id, graph in graphs.items():
# 				print('graph_id: ', graph_id)
# 				print(self._node_maps_from_median[graph_id])
# 				print(self._node_maps_from_median[graph_id].forward_map, self._node_maps_from_median[graph_id].backward_map)
				k = self._node_maps_from_median[graph_id].image(i)
# 				print('k: ', k)
				if k != np.inf:
					node_labels.append(graph.nodes[k])
					
			# Compute the median label and update the median.
			if len(node_labels) > 0:
#				median_label = self._ged_env.get_median_node_label(node_labels)
				median_label = self._get_median_node_label(node_labels)
				if self._ged_env.get_node_rel_cost(median.nodes[i], median_label) > self._epsilon:
					nx.set_node_attributes(median, {i: median_label})
					
					
	def _update_edges(self, graphs, median):
		# Print information about current iteration.
		if self._print_to_stdout == 2:
			print('edges ... ', end='')
			
#		# Clear the adjacency lists of the median and reset number of edges to 0.
# 		median_edges = list(median.edges)		
# 		for (head, tail) in median_edges:
# 			median.remove_edge(head, tail)
		
		# @todo: what if edge is not labeled?
		# Iterate through all possible edges (i,j) of the median.
		for i in range(0, nx.number_of_nodes(median)):
			for j in range(i + 1, nx.number_of_nodes(median)):
				
				# Collect the labels of the edges to which (i,j) is mapped by the node maps.
				edge_labels = []
				for graph_id, graph in graphs.items():
					k = self._node_maps_from_median[graph_id].image(i)
					l = self._node_maps_from_median[graph_id].image(j)
					if k != np.inf and l != np.inf:
						if graph.has_edge(k, l):
							edge_labels.append(graph.edges[(k, l)])
							
				# Compute the median edge label and the overall edge relabeling cost.
				rel_cost = 0
				median_label = self._ged_env.get_edge_label(1)
				if median.has_edge(i, j):
					median_label = median.edges[(i, j)]
				if self._labeled_edges and len(edge_labels) > 0:
					new_median_label = self._get_median_edge_label(edge_labels)
					if self._ged_env.get_edge_rel_cost(median_label, new_median_label) > self._epsilon:
						median_label = new_median_label
					for edge_label in edge_labels:
						rel_cost += self._ged_env.get_edge_rel_cost(median_label, edge_label)
						
				# Update the median.
				if median.has_edge(i, j):
					median.remove_edge(i, j)
				if rel_cost < (self._edge_ins_cost + self._edge_del_cost) * len(edge_labels) - self._edge_del_cost * len(graphs):
					median.add_edge(i, j, **median_label)
# 				else:
# 					if median.has_edge(i, j):
# 						median.remove_edge(i, j)


	def _update_node_maps(self):
		# Update the node maps.
		if self._parallel:
			# @todo: notice when parallel self._ged_env is not modified.
			node_maps_were_modified = False
# 			xxx = self._node_maps_from_median.copy()
			
			len_itr = len(self._node_maps_from_median)
			itr = [item for item in self._node_maps_from_median.items()]
			n_jobs = multiprocessing.cpu_count()
			if len_itr < 100 * n_jobs:
				chunksize = int(len_itr / n_jobs) + 1
			else:
				chunksize = 100
			def init_worker(ged_env_toshare):
				global G_ged_env
				G_ged_env = ged_env_toshare
			nb_nodes_median = self._ged_env.get_graph_num_nodes(self._median_id)
			do_fun = partial(_update_node_maps_parallel, self._median_id, self._epsilon, self._sort_graphs, nb_nodes_median)
			pool = Pool(processes=n_jobs, initializer=init_worker, initargs=(self._ged_env,))
			if self._print_to_stdout == 2:
				iterator = tqdm(pool.imap_unordered(do_fun, itr, chunksize),
							desc='Updating node maps', file=sys.stdout)
			else:
				iterator = pool.imap_unordered(do_fun, itr, chunksize)
			for g_id, node_map, nm_modified in iterator:
				self._node_maps_from_median[g_id] = node_map
				if nm_modified:
					node_maps_were_modified = True
			pool.close()
			pool.join()
# 			yyy = self._node_maps_from_median.copy()

		else:
			# Print information about current iteration.
			if self._print_to_stdout == 2:
				progress = tqdm(desc='Updating node maps', total=len(self._node_maps_from_median), file=sys.stdout)
				
			node_maps_were_modified = False
			nb_nodes_median = self._ged_env.get_graph_num_nodes(self._median_id)
			for graph_id, node_map in self._node_maps_from_median.items():
				nb_nodes_g = self._ged_env.get_graph_num_nodes(graph_id)
				
				if nb_nodes_median <= nb_nodes_g or not self._sort_graphs:
					self._ged_env.run_method(self._median_id, graph_id)
					if self._ged_env.get_upper_bound(self._median_id, graph_id) < node_map.induced_cost() - self._epsilon:
		# 				xxx = self._node_maps_from_median[graph_id]
						self._node_maps_from_median[graph_id] = self._ged_env.get_node_map(self._median_id, graph_id)
						node_maps_were_modified = True
						
				else:
					self._ged_env.run_method(graph_id, self._median_id)
					if self._ged_env.get_upper_bound(graph_id, self._median_id) < node_map.induced_cost() - self._epsilon:
						node_map_tmp = self._ged_env.get_node_map(graph_id, self._median_id)
						node_map_tmp.forward_map, node_map_tmp.backward_map = node_map_tmp.backward_map, node_map_tmp.forward_map
						self._node_maps_from_median[graph_id] = node_map_tmp	
						node_maps_were_modified = True
					
				# Print information about current iteration.
				if self._print_to_stdout == 2:
					progress.update(1)
				
			# Print information about current iteration.
			if self._print_to_stdout == 2:
				print('\n')
			
		# Return true if the node maps were modified.
		return node_maps_were_modified
	
	
	def _decrease_order(self, graphs, median):
		# Print information about current iteration
		if self._print_to_stdout == 2:
			print('Trying to decrease order: ... ', end='')
			
		if nx.number_of_nodes(median) <= 1:
			if self._print_to_stdout == 2:
				print('median graph has only 1 node, skip decrease.')
			return False
			
		# Initialize ID of the node that is to be deleted.
		id_deleted_node = [None] # @todo: or np.inf
		decreased_order = False
		
		# Decrease the order as long as the best deletion delta is negative.
		while self._compute_best_deletion_delta(graphs, median, id_deleted_node) < -self._epsilon:
			decreased_order = True
			self._delete_node_from_median(id_deleted_node[0], median)
			if nx.number_of_nodes(median) <= 1:
				if self._print_to_stdout == 2:
					print('decrease stopped because median graph remains only 1 node. ', end='')
				break
			
		# Print information about current iteration.
		if self._print_to_stdout == 2:
			print('done.')
			
		# Return true iff the order was decreased.
		return decreased_order
	
	
	def _compute_best_deletion_delta(self, graphs, median, id_deleted_node):
		best_delta = 0.0
		
		# Determine node that should be deleted (if any).
		for i in range(0, nx.number_of_nodes(median)):
			# Compute cost delta.
			delta = 0.0
			for graph_id, graph in graphs.items():
				k = self._node_maps_from_median[graph_id].image(i)
				if k == np.inf:
					delta -= self._node_del_cost
				else:
					delta += self._node_ins_cost - self._ged_env.get_node_rel_cost(median.nodes[i], graph.nodes[k])
				for j, j_label in median[i].items():
					l = self._node_maps_from_median[graph_id].image(j)
					if k == np.inf or l == np.inf:
						delta -= self._edge_del_cost
					elif not graph.has_edge(k, l):
						delta -= self._edge_del_cost
					else:
						delta += self._edge_ins_cost - self._ged_env.get_edge_rel_cost(j_label, graph.edges[(k, l)])
						
			# Update best deletion delta.
			if delta < best_delta - self._epsilon:
				best_delta = delta
				id_deleted_node[0] = i
# 			id_deleted_node[0] = 3 # @todo: 
				
		return best_delta
	
	
	def _delete_node_from_median(self, id_deleted_node, median):
		# Update the median.
		mapping = {}
		for i in range(0, nx.number_of_nodes(median)):
			if i != id_deleted_node:
				new_i = (i if i < id_deleted_node else (i - 1))
				mapping[i] = new_i
		median.remove_node(id_deleted_node)
		nx.relabel_nodes(median, mapping, copy=False)
		
		# Update the node maps.
# 		xxx = self._node_maps_from_median
		for key, node_map in self._node_maps_from_median.items():
			new_node_map = NodeMap(nx.number_of_nodes(median), node_map.num_target_nodes())
			is_unassigned_target_node = [True] * node_map.num_target_nodes()
			for i in range(0, nx.number_of_nodes(median) + 1):
				if i != id_deleted_node:
					new_i = (i if i < id_deleted_node else (i - 1))
					k = node_map.image(i)
					new_node_map.add_assignment(new_i, k)
					if k != np.inf:
						is_unassigned_target_node[k] = False
			for k in range(0, node_map.num_target_nodes()):
				if is_unassigned_target_node[k]:
					new_node_map.add_assignment(np.inf, k)
# 			print(self._node_maps_from_median[key].forward_map, self._node_maps_from_median[key].backward_map)
# 			print(new_node_map.forward_map, new_node_map.backward_map
			self._node_maps_from_median[key] = new_node_map
			
		# Increase overall number of decreases.
		self._num_decrease_order += 1
	
	
	def _increase_order(self, graphs, median):
		# Print information about current iteration.
		if self._print_to_stdout == 2:
			print('Trying to increase order: ... ', end='')
			
		# Initialize the best configuration and the best label of the node that is to be inserted.
		best_config = {}
		best_label = self._ged_env.get_node_label(1)
		increased_order = False
		
		# Increase the order as long as the best insertion delta is negative.
		while self._compute_best_insertion_delta(graphs, best_config, best_label) < - self._epsilon:
			increased_order = True
			self._add_node_to_median(best_config, best_label, median)
			
		# Print information about current iteration.
		if self._print_to_stdout == 2:
			print('done.')
			
		# Return true iff the order was increased.
		return increased_order
	
	
	def _compute_best_insertion_delta(self, graphs, best_config, best_label):
		# Construct sets of inserted nodes.
		no_inserted_node = True
		inserted_nodes = {}
		for graph_id, graph in graphs.items():
			inserted_nodes[graph_id] = []
			best_config[graph_id] = np.inf
			for k in range(nx.number_of_nodes(graph)):
				if self._node_maps_from_median[graph_id].pre_image(k) == np.inf:
					no_inserted_node = False
					inserted_nodes[graph_id].append((k, tuple(item for item in graph.nodes[k].items()))) # @todo: can order of label names be garantteed?
					
		# Return 0.0 if no node is inserted in any of the graphs.
		if no_inserted_node:
			return 0.0
		
		# Compute insertion configuration, label, and delta.
		best_delta = 0.0 # @todo
		if len(self._label_names['node_labels']) == 0 and len(self._label_names['node_attrs']) == 0: # @todo
			best_delta = self._compute_insertion_delta_unlabeled(inserted_nodes, best_config, best_label)
		elif len(self._label_names['node_labels']) > 0: # self._constant_node_costs:
			best_delta = self._compute_insertion_delta_constant(inserted_nodes, best_config, best_label)
		else:
			best_delta = self._compute_insertion_delta_generic(inserted_nodes, best_config, best_label)
			
		# Return the best delta.
		return best_delta
	
	
	def _compute_insertion_delta_unlabeled(self, inserted_nodes, best_config, best_label): # @todo: go through and test.
		# Construct the nest configuration and compute its insertion delta.
		best_delta = 0.0
		best_config.clear()
		for graph_id, node_set in inserted_nodes.items():
			if len(node_set) == 0:
				best_config[graph_id] = np.inf
				best_delta += self._node_del_cost
			else:
				best_config[graph_id] = node_set[0][0]
				best_delta -= self._node_ins_cost
				
		# Return the best insertion delta.
		return best_delta
	
	
	def _compute_insertion_delta_constant(self, inserted_nodes, best_config, best_label):
		# Construct histogram and inverse label maps.
		hist = {}
		inverse_label_maps = {}
		for graph_id, node_set in inserted_nodes.items():
			inverse_label_maps[graph_id] = {}
			for node in node_set:
				k = node[0]
				label = node[1]
				if label not in inverse_label_maps[graph_id]:
					inverse_label_maps[graph_id][label] = k
					if label not in hist:
						hist[label] = 1
					else:
						hist[label] += 1
				
		# Determine the best label.
		best_count = 0
		for key, val in hist.items():
			if val > best_count:
				best_count = val
				best_label_tuple = key
		
		# get best label.
		best_label.clear()
		for key, val in best_label_tuple:
			best_label[key] = val
				
		# Construct the best configuration and compute its insertion delta.
		best_config.clear()
		best_delta = 0.0
		node_rel_cost = self._ged_env.get_node_rel_cost(self._ged_env.get_node_label(1), self._ged_env.get_node_label(2))
		triangle_ineq_holds = (node_rel_cost <= self._node_del_cost + self._node_ins_cost)
		for graph_id, _ in inserted_nodes.items():
			if best_label_tuple in inverse_label_maps[graph_id]:
				best_config[graph_id] = inverse_label_maps[graph_id][best_label_tuple]
				best_delta -= self._node_ins_cost
			elif triangle_ineq_holds and not len(inserted_nodes[graph_id]) == 0:
				best_config[graph_id] = inserted_nodes[graph_id][0][0]
				best_delta += node_rel_cost - self._node_ins_cost
			else:
				best_config[graph_id] = np.inf
				best_delta += self._node_del_cost
				
		# Return the best insertion delta.
		return best_delta
	
	
	def _compute_insertion_delta_generic(self, inserted_nodes, best_config, best_label):
		# Collect all node labels of inserted nodes.
		node_labels = []
		for _, node_set in inserted_nodes.items():
			for node in node_set:
				node_labels.append(node[1])
				
		# Compute node label medians that serve as initial solutions for block gradient descent.
		initial_node_labels = []
		self._compute_initial_node_labels(node_labels, initial_node_labels)
		
		# Determine best insertion configuration, label, and delta via parallel block gradient descent from all initial node labels.
		best_delta = 0.0
		for node_label in initial_node_labels:
			# Construct local configuration.
			config = {}
			for graph_id, _ in inserted_nodes.items():
				config[graph_id] = tuple((np.inf, tuple(item for item in self._ged_env.get_node_label(1).items())))
				
			# Run block gradient descent.
			converged = False
			itr = 0
			while not self._insertion_termination_criterion_met(converged, itr):
				converged = not self._update_config(node_label, inserted_nodes, config, node_labels)
				node_label_dict = dict(node_label)
				converged = converged and (not self._update_node_label([dict(item) for item in node_labels], node_label_dict)) # @todo: the dict is tupled again in the function, can be better.
				node_label = tuple(item for item in node_label_dict.items()) # @todo: watch out: initial_node_labels[i] is not modified here.

				itr += 1
				
			# Compute insertion delta of converged solution.
			delta = 0.0
			for _, node in config.items():
				if node[0] == np.inf:
					delta += self._node_del_cost
				else:
					delta += self._ged_env.get_node_rel_cost(dict(node_label), dict(node[1])) - self._node_ins_cost
					
			# Update best delta and global configuration if improvement has been found.
			if delta < best_delta - self._epsilon:
				best_delta = delta
				best_label.clear()
				for key, val in node_label:
					best_label[key] = val
				best_config.clear()
				for graph_id, val in config.items():
					best_config[graph_id] = val[0]
					
		# Return the best delta.
		return best_delta


	def _compute_initial_node_labels(self, node_labels, median_labels):
		median_labels.clear()
		if self._use_real_randomness: # @todo: may not work if parallelized.
			rng = np.random.randint(0, high=2**32 - 1, size=1)
			urng = np.random.RandomState(seed=rng[0])
		else:
			urng = np.random.RandomState(seed=self._seed)
			
		# Generate the initial node label medians.
		if self._init_type_increase_order == 'K-MEANS++':
			# Use k-means++ heuristic to generate the initial node label medians.
			already_selected = [False] * len(node_labels)
			selected_label_id = urng.randint(low=0, high=len(node_labels), size=1)[0] # c++ test: 23
			median_labels.append(node_labels[selected_label_id])
			already_selected[selected_label_id] = True
# 			xxx = [41, 0, 18, 9, 6, 14, 21, 25, 33] for c++ test
# 			iii = 0 for c++ test
			while len(median_labels) < self._num_inits_increase_order:
				weights = [np.inf] * len(node_labels)
				for label_id in range(0, len(node_labels)):
					if already_selected[label_id]:
						weights[label_id] = 0
						continue
					for label in median_labels:
						weights[label_id] = min(weights[label_id], self._ged_env.get_node_rel_cost(dict(label), dict(node_labels[label_id])))
				
				# get non-zero weights.
				weights_p, idx_p = [], []
				for i, w in enumerate(weights):
					if w != 0:
						weights_p.append(w)
						idx_p.append(i)
				if len(weights_p) > 0:
					p = np.array(weights_p) / np.sum(weights_p)
					selected_label_id = urng.choice(range(0, len(weights_p)), size=1, p=p)[0] # for c++ test: xxx[iii] 
					selected_label_id = idx_p[selected_label_id]
# 				iii += 1 for c++ test
					median_labels.append(node_labels[selected_label_id])
					already_selected[selected_label_id] = True
				else: # skip the loop when all node_labels are selected. This happens when len(node_labels) <= self._num_inits_increase_order.
					break
		else:
			# Compute the initial node medians as the medians of randomly generated clusters of (roughly) equal size.
			# @todo: go through and test.
			shuffled_node_labels = [np.inf] * len(node_labels) #@todo: random?
			# @todo: std::shuffle(shuffled_node_labels.begin(), shuffled_node_labels.end(), urng);?
			cluster_size = len(node_labels) / self._num_inits_increase_order
			pos = 0.0
			cluster = []
			while len(median_labels) < self._num_inits_increase_order - 1:
				while pos < (len(median_labels) + 1) * cluster_size:
					cluster.append(shuffled_node_labels[pos])
					pos += 1
				median_labels.append(self._get_median_node_label(cluster))
				cluster.clear()
			while pos < len(shuffled_node_labels):
				pos += 1
				cluster.append(shuffled_node_labels[pos])
			median_labels.append(self._get_median_node_label(cluster))
			cluster.clear()
				
		# Run Lloyd's Algorithm.
		converged = False
		closest_median_ids = [np.inf] * len(node_labels)
		clusters = [[] for _ in range(len(median_labels))]
		itr = 1
		while not self._insertion_termination_criterion_met(converged, itr):
			converged = not self._update_clusters(node_labels, median_labels, closest_median_ids)
			if not converged:
				for cluster in clusters:
					cluster.clear()
				for label_id in range(0, len(node_labels)):
					clusters[closest_median_ids[label_id]].append(node_labels[label_id])
				for cluster_id in range(0, len(clusters)):
					node_label = dict(median_labels[cluster_id])
					self._update_node_label([dict(item) for item in clusters[cluster_id]], node_label) # @todo: the dict is tupled again in the function, can be better.
					median_labels[cluster_id] = tuple(item for item in node_label.items())
			itr += 1
			
			
	def _insertion_termination_criterion_met(self, converged, itr):
		return converged or (itr >= self._max_itrs_increase_order if self._max_itrs_increase_order > 0 else False)
	
	
	def _update_config(self, node_label, inserted_nodes, config, node_labels):
		# Determine the best configuration.
		config_modified = False
		for graph_id, node_set in inserted_nodes.items():
			best_assignment = config[graph_id]
			best_cost = 0.0
			if best_assignment[0] == np.inf:
				best_cost = self._node_del_cost
			else:
				best_cost = self._ged_env.get_node_rel_cost(dict(node_label), dict(best_assignment[1])) - self._node_ins_cost
			for node in node_set:
				cost = self._ged_env.get_node_rel_cost(dict(node_label), dict(node[1])) - self._node_ins_cost
				if cost < best_cost - self._epsilon:
					best_cost = cost
					best_assignment = node
					config_modified = True
			if self._node_del_cost < best_cost - self._epsilon:
				best_cost = self._node_del_cost
				best_assignment = tuple((np.inf, best_assignment[1]))
				config_modified = True
			config[graph_id] = best_assignment
			
		# Collect the node labels contained in the best configuration.
		node_labels.clear()
		for key, val in config.items():
			if val[0] != np.inf:
				node_labels.append(val[1])
		
		# Return true if the configuration was modified.
		return config_modified
				 
	
	def _update_node_label(self, node_labels, node_label):
		if len(node_labels) == 0: # @todo: check if this is the correct solution. Especially after calling _update_config().
			return False
		new_node_label = self._get_median_node_label(node_labels)
		if self._ged_env.get_node_rel_cost(new_node_label, node_label) > self._epsilon:
			node_label.clear()
			for key, val in new_node_label.items():
				node_label[key] = val
			return True
		return False
	
	
	def _update_clusters(self, node_labels, median_labels, closest_median_ids):
		# Determine the closest median for each node label.
		clusters_modified = False
		for label_id in range(0, len(node_labels)):
			closest_median_id = np.inf
			dist_to_closest_median = np.inf
			for median_id in range(0, len(median_labels)):
				dist_to_median = self._ged_env.get_node_rel_cost(dict(median_labels[median_id]), dict(node_labels[label_id]))
				if dist_to_median < dist_to_closest_median - self._epsilon:
					dist_to_closest_median = dist_to_median
					closest_median_id = median_id
			if closest_median_id != closest_median_ids[label_id]:
				closest_median_ids[label_id] = closest_median_id
				clusters_modified = True
				
		# Return true if the clusters were modified.
		return clusters_modified
	
	
	def _add_node_to_median(self, best_config, best_label, median):
		# Update the median.
		nb_nodes_median = nx.number_of_nodes(median)
		median.add_node(nb_nodes_median, **best_label)
		
		# Update the node maps.
		for graph_id, node_map in self._node_maps_from_median.items():
			node_map_as_rel = []
			node_map.as_relation(node_map_as_rel)
			new_node_map = NodeMap(nx.number_of_nodes(median), node_map.num_target_nodes())
			for assignment in node_map_as_rel:
				new_node_map.add_assignment(assignment[0], assignment[1])
			new_node_map.add_assignment(nx.number_of_nodes(median) - 1, best_config[graph_id])
			self._node_maps_from_median[graph_id] = new_node_map
			
		# Increase overall number of increases.
		self._num_increase_order += 1
				
	
	def _are_graphs_equal(self, g1, g2):
		"""
		Check if the two graphs are equal.

		Parameters
		----------
		g1 : NetworkX graph object
			Graph 1 to be compared.
		
		g2 : NetworkX graph object
			Graph 2 to be compared.

		Returns
		-------
		bool
			True if the two graph are equal.
			
		Notes
		-----
		This is not an identical check. Here the two graphs are equal if and only if their original_node_ids, nodes, all node labels, edges and all edge labels are equal. This function is specifically designed for class `MedianGraphEstimator` and should not be used elsewhere.
		"""
		# check original node ids.
		if not g1.graph['original_node_ids'] == g2.graph['original_node_ids']:
			return False
		# check nodes.
		nlist1 = [n for n in g1.nodes(data=True)]
		nlist2 = [n for n in g2.nodes(data=True)]
		if not nlist1 == nlist2:
			return False
		# check edges.
		elist1 = [n for n in g1.edges(data=True)]
		elist2 = [n for n in g2.edges(data=True)]
		if not elist1 == elist2:
			return False

		return True
	
	
	def compute_my_cost(g, h, node_map):
		cost = 0.0
		for node in g.nodes:
			cost += 0
			
			
	def set_label_names(self, node_labels=[], edge_labels=[], node_attrs=[], edge_attrs=[]):
		self._label_names = {'node_labels': node_labels, 'edge_labels': edge_labels,
						'node_attrs': node_attrs, 'edge_attrs': edge_attrs}
			
	
	def _get_median_node_label(self, node_labels):
		if len(self._label_names['node_labels']) > 0:
			return self._get_median_label_symbolic(node_labels)
		elif len(self._label_names['node_attrs']) > 0:
			return self._get_median_label_nonsymbolic(node_labels)
		else:
			raise Exception('Node label names are not given.')
		
			
	def _get_median_edge_label(self, edge_labels):
		if len(self._label_names['edge_labels']) > 0:
			return self._get_median_label_symbolic(edge_labels)
		elif len(self._label_names['edge_attrs']) > 0:
			return self._get_median_label_nonsymbolic(edge_labels)
		else:
			raise Exception('Edge label names are not given.')
			
			
	def _get_median_label_symbolic(self, labels):
		# Construct histogram.
		hist = {}
		for label in labels:
			label = tuple([kv for kv in label.items()]) # @todo: this may be slow.
			if label not in hist:
				hist[label] = 1
			else:
				hist[label] += 1
		
		# Return the label that appears most frequently.
		best_count = 0
		median_label = {}
		for label, count in hist.items():
			if count > best_count:
				best_count = count
				median_label = {kv[0]: kv[1] for kv in label}
				
		return median_label
		
		
	def _get_median_label_nonsymbolic(self, labels):
		if len(labels) == 0:
			return {} # @todo
		else:
			# Transform the labels into coordinates and compute mean label as initial solution.
			labels_as_coords = []
			sums = {}
			for key, val in labels[0].items():
				sums[key] = 0
			for label in labels:
				coords = {}
				for key, val in label.items():
					label_f = float(val)
					sums[key] += label_f
					coords[key] = label_f
				labels_as_coords.append(coords)
			median = {}
			for key, val in sums.items():
				median[key] = val / len(labels)
				
			# Run main loop of Weiszfeld's Algorithm.
			epsilon = 0.0001
			delta = 1.0
			num_itrs = 0
			all_equal = False
			while ((delta > epsilon) and (num_itrs < 100) and (not all_equal)):
				numerator = {}
				for key, val in sums.items():
					numerator[key] = 0
				denominator = 0
				for label_as_coord in labels_as_coords:
					norm = 0
					for key, val in label_as_coord.items():
						norm += (val - median[key]) ** 2
					norm = np.sqrt(norm)
					if norm > 0:
						for key, val in label_as_coord.items():
							numerator[key] += val / norm
						denominator += 1.0 / norm
				if denominator == 0:
					all_equal = True
				else:
					new_median = {}
					delta = 0.0
					for key, val in numerator.items():
						this_median = val / denominator
						new_median[key] = this_median
						delta += np.abs(median[key] - this_median)
					median = new_median
				
				num_itrs += 1
				
			# Transform the solution to strings and return it.
			median_label = {}
			for key, val in median.items():
				median_label[key] = str(val)
			return median_label

		
#	def _get_median_edge_label_symbolic(self, edge_labels):
#		pass
	
	
#	def _get_median_edge_label_nonsymbolic(self, edge_labels):
#		if len(edge_labels) == 0:
#			return {}
#		else:
#			# Transform the labels into coordinates and compute mean label as initial solution.
#			edge_labels_as_coords = []
#			sums = {}
#			for key, val in edge_labels[0].items():
#				sums[key] = 0
#			for edge_label in edge_labels:
#				coords = {}
#				for key, val in edge_label.items():
#					label = float(val)
#					sums[key] += label
#					coords[key] = label
#				edge_labels_as_coords.append(coords)
#			median = {}
#			for key, val in sums.items():
#				median[key] = val / len(edge_labels)
#				
#			# Run main loop of Weiszfeld's Algorithm.
#			epsilon = 0.0001
#			delta = 1.0
#			num_itrs = 0
#			all_equal = False
#			while ((delta > epsilon) and (num_itrs < 100) and (not all_equal)):
#				numerator = {}
#				for key, val in sums.items():
#					numerator[key] = 0
#				denominator = 0
#				for edge_label_as_coord in edge_labels_as_coords:
#					norm = 0
#					for key, val in edge_label_as_coord.items():
#						norm += (val - median[key]) ** 2
#					norm += np.sqrt(norm)
#					if norm > 0:
#						for key, val in edge_label_as_coord.items():
#							numerator[key] += val / norm
#						denominator += 1.0 / norm
#				if denominator == 0:
#					all_equal = True
#				else:
#					new_median = {}
#					delta = 0.0
#					for key, val in numerator.items():
#						this_median = val / denominator
#						new_median[key] = this_median
#						delta += np.abs(median[key] - this_median)
#					median = new_median
#					
#				num_itrs += 1
#				
#			# Transform the solution to ged::GXLLabel and return it.
#			median_label = {}
#			for key, val in median.items():
#				median_label[key] = str(val)
#			return median_label


def _compute_medoid_parallel(graph_ids, sort, itr):
	g_id = itr[0]
	i = itr[1]
	# @todo: timer not considered here.
# 			if timer.expired():
# 				self._state = AlgorithmState.CALLED
# 				break
	nb_nodes_g = G_ged_env.get_graph_num_nodes(g_id)
	sum_of_distances = 0
	for h_id in graph_ids:
		nb_nodes_h = G_ged_env.get_graph_num_nodes(h_id)
		if nb_nodes_g <= nb_nodes_h or not sort:
			G_ged_env.run_method(g_id, h_id)
			sum_of_distances += G_ged_env.get_upper_bound(g_id, h_id)
		else:
			G_ged_env.run_method(h_id, g_id)
			sum_of_distances += G_ged_env.get_upper_bound(h_id, g_id)
	return i, sum_of_distances
				

def _compute_init_node_maps_parallel(gen_median_id, sort, nb_nodes_median, itr):
	graph_id = itr
	nb_nodes_g = G_ged_env.get_graph_num_nodes(graph_id)
	if nb_nodes_median <= nb_nodes_g or not sort:
		G_ged_env.run_method(gen_median_id, graph_id)
		node_map = G_ged_env.get_node_map(gen_median_id, graph_id)
# 				print(self._node_maps_from_median[graph_id])
	else:
		G_ged_env.run_method(graph_id, gen_median_id)
		node_map = G_ged_env.get_node_map(graph_id, gen_median_id)
		node_map.forward_map, node_map.backward_map = node_map.backward_map, node_map.forward_map
	sum_of_distance = node_map.induced_cost()
# 				print(self._sum_of_distances)
	return graph_id, sum_of_distance, node_map
					

def _update_node_maps_parallel(median_id, epsilon, sort, nb_nodes_median, itr):
	graph_id = itr[0]
	node_map = itr[1]

	node_maps_were_modified = False
	nb_nodes_g = G_ged_env.get_graph_num_nodes(graph_id)
	if nb_nodes_median <= nb_nodes_g or not sort:
		G_ged_env.run_method(median_id, graph_id)
		if G_ged_env.get_upper_bound(median_id, graph_id) < node_map.induced_cost() - epsilon:
			node_map = G_ged_env.get_node_map(median_id, graph_id)
			node_maps_were_modified = True			
	else:
		G_ged_env.run_method(graph_id, median_id)
		if G_ged_env.get_upper_bound(graph_id, median_id) < node_map.induced_cost() - epsilon:
			node_map = G_ged_env.get_node_map(graph_id, median_id)
			node_map.forward_map, node_map.backward_map = node_map.backward_map, node_map.forward_map
			node_maps_were_modified = True	
			
	return graph_id, node_map, node_maps_were_modified


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun 17 17:49:24 2020

@author: ljia
"""


class EditCost(object):
	
	
	def __init__(self):
		pass
	
	
	def node_ins_cost_fun(self, node_label):
		"""
	/*!
	 * @brief Node insertions cost function.
	 * @param[in] node_label A node label.
	 * @return The cost of inserting a node with label @p node_label.
	 * @note Must be implemented by derived classes of ged::EditCosts.
	 */		
		"""
		return 0

	
	def node_del_cost_fun(self, node_label):
		"""
	/*!
	 * @brief Node deletion cost function.
	 * @param[in] node_label A node label.
	 * @return The cost of deleting a node with label @p node_label.
	 * @note Must be implemented by derived classes of ged::EditCosts.
	 */		
		"""
		return 0

	
	def node_rel_cost_fun(self, node_label_1, node_label_2):
		"""
	/*!
	 * @brief Node relabeling cost function.
	 * @param[in] node_label_1 A node label.
	 * @param[in] node_label_2 A node label.
	 * @return The cost of changing a node's label from @p node_label_1 to @p node_label_2.
	 * @note Must be implemented by derived classes of ged::EditCosts.
	 */		
		"""
		return 0

	
	def edge_ins_cost_fun(self, edge_label):
		"""
	/*!
	 * @brief Edge insertion cost function.
	 * @param[in] edge_label An edge label.
	 * @return The cost of inserting an edge with label @p edge_label.
	 * @note Must be implemented by derived classes of ged::EditCosts.
	 */		
		"""
		return 0

	
	def edge_del_cost_fun(self, edge_label):
		"""
	/*!
	 * @brief Edge deletion cost function.
	 * @param[in] edge_label An edge label.
	 * @return The cost of deleting an edge with label @p edge_label.
	 * @note Must be implemented by derived classes of ged::EditCosts.
	 */		
		"""
		return 0

	
	def edge_rel_cost_fun(self, edge_label_1, edge_label_2):
		"""
	/*!
	 * @brief Edge relabeling cost function.
	 * @param[in] edge_label_1 An edge label.
	 * @param[in] edge_label_2 An edge label.
	 * @return The cost of changing an edge's label from @p edge_label_1 to @p edge_label_2.
	 * @note Must be implemented by derived classes of ged::EditCosts.
	 */		
		"""
		return 0


"""
util2

GED computation utilities for the methods that can learn different edit costs
for each pair of (symbolic) labels.

@Author: linlin
@Date: 11.05.23
"""
import networkx as nx
import numpy as np

from gklearn.ged.util.util import ged_options_to_string


def pairwise_ged2(
		g1, g2, options={}, sort=True, repeats=1, parallel=False, verbose=True
):
	"""Compute the graph edit distance between two graphs using the gedlib library
	with repeats. This method is used for the methods that can learn different
	edit costs for each pair of (symbolic) labels.

	Notes
	-----
		- For methods such as BIPARTITE, the repeats may result same results.
		- # of edit operations are not computed in this method.
	"""
	from gklearn.gedlib import librariesImport, gedlibpy2

	ged_env = gedlibpy2.GEDEnv()
	ged_env.set_edit_cost(
		options['edit_cost'],
		edit_cost_constant=options['edit_cost_constants']
	)
	ged_env.add_nx_graph(g1, '')
	ged_env.add_nx_graph(g2, '')
	listID = ged_env.get_all_graph_ids()
	ged_env.init(
		init_option=(options[
			             'init_option'] if 'init_option' in options else 'EAGER_WITHOUT_SHUFFLED_COPIES')
	)
	ged_env.set_method(options['method'], ged_options_to_string(options))
	ged_env.init_method()

	g = listID[0]
	h = listID[1]
	dis_min = np.inf
	# 	print('------------------------------------------')
	for i in range(0, repeats):
		ged_env.run_method(g, h)
		upper = ged_env.get_upper_bound(g, h)
		dis = upper
		# 		print(dis)
		if dis < dis_min:
			dis_min = dis
			pi_forward = ged_env.get_forward_map(g, h)
			pi_backward = ged_env.get_backward_map(g, h)
	# 			lower = ged_env.get_lower_bound(g, h)

	# make the map label correct (label remove map as np.inf)
	nodes1 = [n for n in g1.nodes()]
	nodes2 = [n for n in g2.nodes()]
	nb1 = nx.number_of_nodes(g1)
	nb2 = nx.number_of_nodes(g2)
	pi_forward = [nodes2[pi] if pi < nb2 else np.inf for pi in pi_forward]
	pi_backward = [nodes1[pi] if pi < nb1 else np.inf for pi in pi_backward]
	#		print(pi_forward)

	return dis, pi_forward, pi_backward


# %%


def get_nb_edit_operations2(
		g1, g2, forward_map, backward_map, edit_cost=None, is_cml=False,
		**kwargs
):
	"""Calculate the numbers of the occurence of each edit operation in a given
	edit path.

	Parameters
	----------
	g1 : TYPE
		DESCRIPTION.
	g2 : TYPE
		DESCRIPTION.
	forward_map : TYPE
		DESCRIPTION.
	backward_map : TYPE
		DESCRIPTION.
	edit_cost : TYPE, optional
		DESCRIPTION. The default is None.
	is_cml : TYPE, optional
		DESCRIPTION. The default is False.
	**kwargs : TYPE
		DESCRIPTION.

	Raises
	------
	Exception
		DESCRIPTION.

	Returns
	-------
	TYPE
		DESCRIPTION.

	Notes
	-----
	Attention: when implementing a function to get the numbers of edit
	operations, make sure that:
		- It does not fail if NetworkX node labels are not consecutive integers;
		- It returns correct results if nodes are permutated (e.g., by using
		`gklearn.utis.utils.nx_permute_nodes()`.)
	Generally speaking, it means you need to distinguish the NetworkX label of
	a node from the position (index) of that node in the node list.
	"""
	if is_cml:
		if edit_cost == 'CONSTANT':
			node_labels = kwargs.get('node_labels', [])
			edge_labels = kwargs.get('edge_labels', [])
			return get_nb_edit_operations_symbolic_cml(
				g1, g2, forward_map, backward_map,
				node_labels=node_labels, edge_labels=edge_labels
			)
		else:
			raise Exception('Edit cost "', edit_cost, '" is not supported.')
	else:
		if edit_cost == 'LETTER' or edit_cost == 'LETTER2':
			return get_nb_edit_operations_letter(
				g1, g2, forward_map,
				backward_map
			)
		elif edit_cost == 'NON_SYMBOLIC':
			# Use node_labels and edge_labels if node_attrs and edge_attrs are
			# not specified. Otherwise, set them to [].
			node_attrs = kwargs.get('node_attrs', kwargs.get('node_labels', []))
			edge_attrs = kwargs.get('edge_attrs', kwargs.get('edge_labels', []))
			return get_nb_edit_operations_nonsymbolic(
				g1, g2, forward_map, backward_map,
				node_attrs=node_attrs, edge_attrs=edge_attrs
			)
		elif edit_cost == 'CONSTANT':
			# node_labels = kwargs.get('node_labels', [])
			# edge_labels = kwargs.get('edge_labels', [])
			edit_cost_constants = kwargs.get('edit_cost_constants')
			return_sorted_costs = kwargs.get('return_sorted_costs')
			return get_nb_edit_operations_symbolic2(
				g1, g2, forward_map, backward_map,
				node_labels='l', edge_labels='l',  # TODO: to generalize.
				edit_cost_constants=edit_cost_constants,
				return_sorted_costs=return_sorted_costs
			)
		else:
			raise Exception('Edit cost "', edit_cost, '" is not supported.')
			return get_nb_edit_operations_symbolic2(
				g1, g2, forward_map,
				backward_map
			)


def get_nb_edit_operations_symbolic2(
		g1, g2, forward_map, backward_map, node_labels='l', edge_labels='l',
		edit_cost_constants=None, return_sorted_costs=True
):
	"""Compute the number of each edit operations for symbolic-labeled graphs,
	where the edit costs are different costants for different labels.

	The node labels and edge labels of the whole dataset are first obtained from
	`edit_cost_constants` and then ordered in lexicographical order.
	Then, the numbers of edit operations is computed for each pair of labels.
	The returned numebers of edit operations is a list in the following order:
		- number of node insertion for each pair of labels;
		- number of node deletion for each pair of labels;
		- number of node substitution for each pair of labels;
		- number of edge insertion for each pair of labels;
		- number of edge deletion for each pair of labels;
		- number of edge substitution for each pair of labels.
	if the nodes or edges are not labeled, the corresponding numbers of 3 edit
	operations for unified edit costs are computed, namely, the costs of insertion,
	deletion, and substitution.

	Parameters
	----------
	g1 : networkx.Graph
		The first graph.
	g2 : networkx.Graph
		The second graph.
	forward_map : list
		The forward map.
	backward_map : list
		The backward map.
	node_labels : str, optional
		The node label attribute name. The default is 'l'.
	edge_labels : str, optional
		The edge label attribute name. The default is 'l'.
	edit_cost_constants : tuple(dict, dict), optional
		The edit cost constants. A tuple of two dictionaries, the first one for
		node labels and the second one for edge labels. The default is None.
	return_sorted_costs : bool, optional
		Whether to return the sorted edit costs. The default is True.

	Returns
	-------
	nb_edit_operations : list
		The numbers of edit operations for each pair of labels.
	sorted_label_pairs : tuple(list, list)
		The sorted label pairs for node labels and edge labels, respectively.
	sorted_edit_costs : list, optional
		The sorted edit costs, first for node labels and then for edge labels.
		if `return_sorted_costs` is True.

	Notes
	-----
	`node_labels` and `edge_labels` only support 'l' for now.

	Authors
	-------
	Linlin Jia, GitHub Copilot (2023.05.13)
	"""
	# Get the node labels and edge labels of the whole dataset.
	if edit_cost_constants is None:
		raise Exception('Edit cost constants are not specified.')
	ori_node_labels = _get_labels_from_edit_cost_const_dict(
		edit_cost_constants[0]
	)
	ori_edge_labels = _get_labels_from_edit_cost_const_dict(
		edit_cost_constants[1]
	)
	# Order the labels and get the label pairs, including the special label pairs
	# for insertion and deletion.
	nl_pairs = _get_ordered_label_pairs(ori_node_labels)
	el_pairs = _get_ordered_label_pairs(ori_edge_labels)

	# Iterate over the forward map and backward map to compute the numbers of edit
	# operations.
	# for nodes:
	if len(ori_node_labels) == 0:
		# if nodes are not labeled:
		n_node_ops = _get_n_node_ops_unified(
			forward_map, backward_map, g1, g2, node_labels
		)
	else:
		# if nodes are labeled:
		n_node_ops = _get_n_node_ops_symbolic(
			forward_map, backward_map, g1, g2, node_labels, nl_pairs,
			edit_cost_constants[0], len(ori_node_labels)
		)

	# for edges:
	if len(ori_edge_labels) == 0:
		# if edges are not labeled:
		n_edge_ops = _get_n_edge_ops_unified(
			forward_map, backward_map, g1, g2, edge_labels
		)
	else:
		# if edges are labeled:
		n_edge_ops = _get_n_edge_ops_symbolic(
			forward_map, backward_map, g1, g2, edge_labels, el_pairs,
			edit_cost_constants[1], len(ori_edge_labels)
		)

	if return_sorted_costs:
		sorted_costs_n = sort_edit_costs_by_keys(
			edit_cost_constants[0], nl_pairs, len(ori_node_labels)
		)
		sorted_costs_e = sort_edit_costs_by_keys(
			edit_cost_constants[1], el_pairs, len(ori_edge_labels)
		)
		return n_node_ops + n_edge_ops, (
		nl_pairs, el_pairs), sorted_costs_n + sorted_costs_e
	else:
		return n_node_ops + n_edge_ops, (nl_pairs, el_pairs)


def _get_n_node_ops_unified(
		forward_map, backward_map, g1, g2, node_labels
):
	"""
	Compute the numbers of node insertion, removal, and substitution for unified
	edit costs.

	Parameters
	----------
	forward_map : list
		The forward map.
	backward_map : list
		The backward map.
	g1 : networkx.Graph
		The first graph.
	g2 : networkx.Graph
		The second graph.
	node_labels : str
		The node label attribute name.

	Returns
	-------
	list of int
		The numbers of node insertion, removal, and substitution.

	Notes
	-----
	`node_labels` only supports 'l' for now.
	"""
	n_vi = 0
	n_vr = 0
	n_vs = 0

	nodes1 = [n for n in g1.nodes()]
	for i, map_i in enumerate(forward_map):
		if map_i == np.inf:
			n_vr += 1
		else:
			continue  # @TODO: fo now only support unlabeled edges.
			# The nodes are considered substituted as long as at least one feature
			# is different.
			label1 = g1.nodes[nodes1[i]]['l']
			label2 = g2.nodes[map_i]['l']
			if label1 != label2:
				n_vs += 1
	for map_i in backward_map:
		if map_i == np.inf:
			n_vi += 1
	n_ops = [n_vi, n_vr, n_vs]
	return n_ops


# n_vi = 0
# n_vr = 0
# n_vs = 0
#
# nodes1 = [n for n in g1.nodes()]
# for i, map_i in enumerate(forward_map):
# 	if map_i == np.inf:
# 		n_vr += 1
# 	else:
# 		# The nodes are considered substituted as long as at least one feature
# 		# is different.
# 		for nl in node_labels:
# 			label1 = g1.nodes[nodes1[i]][nl]
# 			label2 = g2.nodes[map_i][nl]
# 			if label1 != label2:
# 				n_vs += 1
# 				break
# for map_i in backward_map:
# 	if map_i == np.inf:
# 		n_vi += 1
# return n_vi, n_vr, n_vs, nodes1


def _get_n_node_ops_symbolic(
		forward_map, backward_map, g1, g2, node_labels, nl_pairs, edit_costs,
		n_ori_node_labels
):
	"""
	Compute the numbers of node insertion, removal, and substitution for symbolic
	edit costs.

	Parameters
	----------
	forward_map : list
		The forward map.
	backward_map : list
		The backward map.
	g1 : networkx.Graph
		The first graph.
	g2 : networkx.Graph
		The second graph.
	node_labels : str
		The node label attribute name.
	nl_pairs : list
		The ordered node label pairs.
	edit_costs : dict
		The edit costs. The keys are the node label pairs. The values are the edit
		costs.

	Returns
	-------
	list of int
		The numbers of node insertion, removal, and substitution.

	Notes
	-----
	`node_labels` only supports 'l' for now.

	Authors
	-------
	Linlin Jia, GitHub Copilot (2023.05.13)
	"""
	# Initialize the numbers of edit operations as 0.
	n_node_ops = [0] * len(nl_pairs)

	nodes1 = [n for n in g1.nodes()]
	# Iterate over the forward map:
	# @TODO: check if map_i the indices of the nodes in g2 or the real nx labels of the nodes.
	for i, map_i in enumerate(forward_map):
		if map_i == np.inf:
			# The node is removed:
			# Get the node label.
			label1 = g1.nodes[nodes1[i]]['l']
			# Find the index of the node label in the ordered node label pairs.
			idx = nl_pairs.index((label1, ''))
			# Increment the number of node removals.
			n_node_ops[idx] += 1
		else:
			# The node is substituted:
			# The nodes are considered substituted as long as at least one feature
			# is different.
			label1 = g1.nodes[nodes1[i]]['l']
			label2 = g2.nodes[map_i]['l']
			if label1 != label2:
				# Find the index of the node label pair in the ordered node label pairs.
				idx = nl_pairs.index((label1, label2))
				# Increment the number of node substitutions.
				n_node_ops[idx] += 1
	# Iterate over the backward map to find node insertions:
	nodes2 = [n for n in g2.nodes()]
	for i, map_i in enumerate(backward_map):
		if map_i == np.inf:
			# The node is inserted:
			# Get the node label.
			label2 = g2.nodes[nodes2[i]]['l']
			# Find the index of the node label in the ordered node label pairs.
			idx = nl_pairs.index(('', label2))
			# Increment the number of node insertions.
			n_node_ops[idx] += 1

	return n_node_ops


def _get_n_edge_ops_unified(
		forward_map, backward_map, g1, g2, edge_labels
):
	"""
	Compute the numbers of edge insertion, removal, and substitution for unified
	edit costs.

	Parameters
	----------
	forward_map : list
		The forward map.
	backward_map : list
		The backward map.
	g1 : networkx.Graph
		The first graph.
	g2 : networkx.Graph
		The second graph.
	edge_labels : str
		The edge label attribute name.

	Returns
	-------
	list of int
		The numbers of edge insertion, removal, and substitution.

	Notes
	-----
	`edge_labels` only supports 'l' for now.

	Authors
	-------
	Linlin Jia, GitHub Copilot (2023.05.13)
	"""
	n_ei = 0
	n_er = 0
	n_es = 0

	nodes1 = [n for n in g1.nodes()]
	edges1 = [e for e in g1.edges()]
	nb_edges2_cnted = 0
	for n1, n2 in edges1:
		idx1 = nodes1.index(n1)
		idx2 = nodes1.index(n2)
		# One of the nodes is removed, thus the edge is removed.
		if forward_map[idx1] == np.inf or forward_map[idx2] == np.inf:
			n_er += 1
		# Corresponding edge is in g2.
		# Notice if the graph is undirected, the two directions of the edge can
		# both be found in the nx.edges() list, so the next elif will not be
		# executed; meanwhile, nx.number_of_edges() will count the two directions
		# only once. nb_edges2_cnted is thus correct.
		elif (forward_map[idx1], forward_map[idx2]) in g2.edges():
			nb_edges2_cnted += 1
			continue  # @TODO: fo now only support unlabeled edges.
			# Edge labels are different as long as any one feature is different.
			el = 'l'
			label1 = g2.edges[((forward_map[idx1], forward_map[idx2]))][el]
			label2 = g1.edges[(n1, n2)][el]
			if label1 != label2:
				n_es += 1
		elif (forward_map[idx2], forward_map[idx1]) in g2.edges():
			nb_edges2_cnted += 1
			continue  # @TODO: fo now only support unlabeled edges.
			# Edge labels are different.
			el = 'l'
			label1 = g2.edges[((forward_map[idx2], forward_map[idx1]))][el]
			label2 = g1.edges[(n1, n2)][el]
			if label1 != label2:
				n_es += 1
		# corresponding nodes are in g2, however the edge is removed.
		else:
			n_er += 1
	n_ei = nx.number_of_edges(g2) - nb_edges2_cnted

	n_ops = [n_ei, n_er, n_es]

	return n_ops


# n_ei = 0
# n_er = 0
# n_es = 0
#
# edges1 = [e for e in g1.edges()]
# nb_edges2_cnted = 0
# for n1, n2 in edges1:
# 	idx1 = nodes1.index(n1)
# 	idx2 = nodes1.index(n2)
# 	# One of the nodes is removed, thus the edge is removed.
# 	if forward_map[idx1] == np.inf or forward_map[idx2] == np.inf:
# 		n_er += 1
# 	# Corresponding edge is in g2.
# 	elif (forward_map[idx1], forward_map[idx2]) in g2.edges():
# 		nb_edges2_cnted += 1
# 		# Edge labels are different as long as any one feature is different.
# 		for el in edge_labels:
# 			label1 = g2.edges[((forward_map[idx1], forward_map[idx2]))][el]
# 			label2 = g1.edges[(n1, n2)][el]
# 			if label1 != label2:
# 				n_es += 1
# 				break
# 	elif (forward_map[idx2], forward_map[idx1]) in g2.edges():
# 		nb_edges2_cnted += 1
# 		# Edge labels are different.
# 		for el in edge_labels:
# 			label1 = g2.edges[((forward_map[idx2], forward_map[idx1]))][el]
# 			label2 = g1.edges[(n1, n2)][el]
# 			if label1 != label2:
# 				n_es += 1
# 				break
# 	# corresponding nodes are in g2, however the edge is removed.
# 	else:
# 		n_er += 1
# n_ei = nx.number_of_edges(g2) - nb_edges2_cnted
#
# return n_vi, n_vr, n_vs, n_ei, n_er, n_es


def _get_n_edge_ops_symbolic(
		forward_map,
		backward_map,
		g1,
		g2,
		edge_labels,
		el_pairs,
		edit_costs,
		n_ori_edges_labels
):
	"""
	Compute the numbers of edge insertion, removal, and substitution for symbolic
	edit costs.

	Parameters
	----------
	forward_map : list
		The forward map.
	backward_map : list
		The backward map.
	g1 : networkx.Graph
		The first graph.
	g2 : networkx.Graph
		The second graph.
	edge_labels : str
		The edge label attribute name.
	el_pairs : list
		The ordered edge label pairs.
	edit_costs : dict
		The edit costs. The keys are the edge label pairs. The values are the edit
		costs.

	Returns
	-------
	list of int
		The numbers of edge insertion, removal, and substitution.

	Notes
	-----
		- `edge_labels` only supports 'l' for now.
		- This function only supports undirected graphs for now.

	Authors
	-------
	Linlin Jia, GitHub Copilot (2023.05.13)
	"""
	# Notice if the graph is undirected, list of nx.edges() only returns one
	# direction of the edges, but the two directions of the edge can
	# both be found in the nx.edges() list; meanwhile, nx.number_of_edges() will
	# count the two directions only once. # @TODO

	# Initialize the numbers of edit operations as 0.
	n_edit_ops = [0] * len(el_pairs)

	nodes1 = [n for n in g1.nodes()]
	edges1 = [e for e in g1.edges()]
	edges2_cnted = []  # Edges in g2 that have been counted.
	# Iterate over the edges in g1.
	for n1, n2 in edges1:
		idx1 = nodes1.index(n1)
		idx2 = nodes1.index(n2)
		# One of the nodes is removed, thus the edge is removed.
		if forward_map[idx1] == np.inf or forward_map[idx2] == np.inf:
			# Get the edge label.
			label1 = g1.edges[(n1, n2)]['l']
			# Get the index of the edge label pair.
			idx = el_pairs.index((label1, ''))
			# Increment the number of edge removals.
			n_edit_ops[idx] += 1
		# Corresponding edge is in g2.
		elif (forward_map[idx1], forward_map[idx2]) in g2.edges():
			# Add the edge to the counted list.
			edges2_cnted.append((forward_map[idx1], forward_map[idx2]))
			# Edge labels are different as long as any one feature is different.
			el = 'l'  #  @TODO
			label2 = g2.edges[((forward_map[idx1], forward_map[idx2]))][el]
			label1 = g1.edges[(n1, n2)][
				el]  # todo: check if the order of label1 and label2 is correct.
			if label1 != label2:
				# Get the index of the edge label pair.
				idx = el_pairs.index((label1, label2))
				# Increment the number of edge substitutions.
				n_edit_ops[idx] += 1
		# This `elif` will never be reached if the edges are undirected.
		elif (forward_map[idx2], forward_map[idx1]) in g2.edges():
			# Add the edge to the counted list.
			edges2_cnted.append((forward_map[idx2], forward_map[idx1]))
			# Edge labels are different.
			el = 'l'
			label2 = g2.edges[((forward_map[idx2], forward_map[idx1]))][el]
			label1 = g1.edges[(n1, n2)][el]
			if label1 != label2:
				# Get the index of the edge label pair.
				idx = el_pairs.index((label1, label2))
				# Increment the number of edge substitutions.
				n_edit_ops[idx] += 1
		# corresponding nodes are in g2, however the edge is removed.
		else:
			# Get the edge label.
			label1 = g1.edges[(n1, n2)]['l']
			# Get the index of the edge label pair.
			idx = el_pairs.index((label1, ''))
			# Increment the number of edge removals.
			n_edit_ops[idx] += 1

	# Iterate over the edges in g2 to find the inserted edges:
	edges2 = [e for e in g2.edges()]
	for n1, n2 in edges2:
		# idx1 = nodes1.index(n1)
		# idx2 = nodes1.index(n2)
		# Check if the edge is already counted:
		if nx.is_directed(g2):
			if (n1, n2) in edges2_cnted:
				continue
		else:
			if (n1, n2) in edges2_cnted or (n2, n1) in edges2_cnted:
				continue

		# Get the edge label.
		label2 = g2.edges[(n1, n2)]['l']
		# Get the index of the edge label pair.
		idx = el_pairs.index(('', label2))
		# Increment the number of edge insertions.
		n_edit_ops[idx] += 1

	return n_edit_ops


# %%


def _retrieve_ordered_label_pairs_from_cost_dict(cost_dict):
	"""Get the ordered label pairs from the keys of edit cost constants.

	Parameters
	----------
	cost_dict : dict
		The edit cost constants, where the keys are the tuples of labels.

	Returns
	-------
	list
		The ordered label pairs.

	Notes
	-----
	The special label '' is included for insertion and deletion; the special label
	' ' is included for unlabeled nodes or edges.

	Authors
	-------
	Linlin Jia, GitHub Copilot (2023.05.13)
	"""
	# Get the node labels and edge labels of the whole dataset.
	ori_labels = _get_labels_from_edit_cost_const_dict(cost_dict)

	# if the nodes or edges are not labeled, the special label for 3 edit operations
	# is returned:
	if len(ori_labels) == 0:
		return [('', ' '), (' ', ''), (' ', ' ')]

	# if the nodes or edges are labeled, the label pairs are returned:
	label_paris = get_label_pairs(ori_labels, sort=True, extended=True)
	return label_paris


def _get_ordered_label_pairs(ori_labels):
	"""Get the ordered label pairs from the original labels.

	Parameters
	----------
	ori_labels : list
		The original labels.

	Returns
	-------
	list
		The ordered label pairs.

	Notes
	-----
	The special label '' is included for insertion and deletion; the special label
	' ' is included for unlabeled nodes or edges.

	Authors
	-------
	Linlin Jia, GitHub Copilot (2023.05.13)
	"""
	# if the nodes or edges are not labeled, the special label for 3 edit operations
	# is returned:
	if len(ori_labels) == 0:
		return [('', ' '), (' ', ''), (' ', ' ')]

	# if the nodes or edges are labeled, the label pairs are returned:
	label_paris = get_label_pairs(ori_labels, sort=True, extended=True)
	return label_paris


def _get_labels_from_edit_cost_const_dict(cost_dict):
	"""Get the node labels or edge labels from the keys of edit cost constants.

	Parameters
	----------
	cost_dict : dict
		The edit cost constants, where the keys are the tuples of labels.

	Returns
	-------
	list
		The node labels or edge labels.

	Authors
	-------
	Linlin Jia, GitHub Copilot (2023.05.13)
	"""
	labels = []
	for k in cost_dict.keys():
		for l in k:
			if l not in labels:
				labels.append(l)
	# Remove special labels: '', ' ':
	if '' in labels:
		labels.remove('')
	if ' ' in labels:
		labels.remove(' ')
	return labels


def sort_edit_costs_by_keys(costs_dict, ordered_label_pairs, n_ori_labels):
	"""
	Sort the edit costs by the keys.

	Parameters
	----------
	costs_dict : dict
		The edit costs, where the keys are the tuples of labels.
	ordered_label_pairs : list
		The ordered label pairs.
	n_ori_labels : int
		The number of original labels.

	Returns
	-------
	list
		The sorted edit costs.

	Authors
	-------
	Linlin Jia, GitHub Copilot (2023.05.14)
	"""
	if n_ori_labels == 0:
		# for unlabeled nodes or edges:
		return [costs_dict[('', ' ')], costs_dict[(' ', '')],
		        costs_dict[(' ', ' ')]]
	else:
		# for labeled nodes or edges:
		costs = []
		for l1, l2 in ordered_label_pairs:
			costs.append(costs_dict[(l1, l2)])
		return costs


def get_label_pairs(labels, sort=True, extended=False):
	"""
	Get all possible pairs of labels from a list of labels.

	Parameters
	----------
	labels : list of hashable
		List of labels.
	sort : bool, optional (default=True)
		Sort labels to ensure that the same pair of labels is always represented.
	extended : bool, optional (default=False)
		Include insertions and deletions pairs.

	Returns
	-------
	label_pairs : list of tuples
		List of all possible pairs of labels.
	"""
	# Sort labels to ensure that the same pair of labels is always represented
	# by the same tuple.
	if sort:
		labels = sorted(labels)

	label_pairs = []

	# Include insertions and deletions pairs.
	if extended:
		for i in range(len(labels)):
			if isinstance(i, str) and i == '':
				raise ValueError(
					'The empty string is not allowed as a label. As it is used '
					'to represent the special label (epsilon) for insertion and deletion.'
				)
			label_pairs.append(('', labels[i]))
		for i in range(len(labels)):
			label_pairs.append((labels[i], ''))

	# Get all possible pairs of labels, two directions.
	for i in range(len(labels)):
		for j in range(i + 1, len(labels)):
			label_pairs.append((labels[i], labels[j]))
			label_pairs.append((labels[j], labels[i]))

	return label_pairs


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Mar 30 11:52:47 2020

@author: ljia
"""
import time
import functools
import multiprocessing

import numpy as np
import networkx as nx
# from abc import ABC, abstractmethod

from sklearn.base import BaseEstimator  # , TransformerMixin
from sklearn.utils.validation import check_is_fitted  # check_X_y, check_array,
from sklearn.exceptions import NotFittedError

from gklearn.utils import normalize_gram_matrix, is_basic_python_type


class GraphKernel(BaseEstimator):  # , ABC):
	"""The basic graph kernel class.

	Attributes
	----------
	_graphs : list
		Stores the input graphs on fit input data.
		Default format of the list objects is `NetworkX` graphs.
		**We don't guarantee that the input graphs remain unchanged during the
		computation.**

	References
	----------
	https://ysig.github.io/GraKeL/0.1a8/_modules/grakel/kernels/kernel.html#Kernel.
	"""


	def __init__(
			self,
			parallel=None,
			n_jobs=None,
			chunksize=None,
			normalize=True,
			copy_graphs=True,  # make sure it is a full deep copy. and faster!
			verbose=2
	):
		"""`__init__` for `GraphKernel` object."""
		# @todo: the default settings of the parameters are different from those in the self.compute method.
		# 		self._graphs = None
		self.parallel = parallel
		self.n_jobs = n_jobs
		self.chunksize = chunksize
		self.normalize = normalize
		self.copy_graphs = copy_graphs
		self.verbose = verbose


	# 		self._run_time = 0
	# 		self._gram_matrix = None
	# 		self._gram_matrix_unnorm = None

	##########################################################################
	# The following is the 1st paradigm to compute kernel matrix, which is
	# compatible with `scikit-learn`.
	# -------------------------------------------------------------------
	# Special thanks to the "GraKeL" library for providing an excellent template!
	##########################################################################

	def fit(self, X, y=None):
		"""Fit a graph dataset for a transformer.

		Parameters
		----------
		X : iterable
			DESCRIPTION.

		y : None, optional
			There is no need of a target in a transformer, yet the `scikit-learn`
			pipeline API requires this parameter.

		Returns
		-------
		object
			Returns self.

		"""
		# 		self._is_tranformed = False

		# Clear any prior attributes stored on the estimator, # @todo: unless warm_start is used;
		self.clear_attributes()

		# Validate parameters for the transformer.
		self.validate_parameters()

		# Validate the input.
		self._graphs = self.validate_input(X)

		# 		self._X = X
		# 		self._kernel = self._get_kernel_instance()

		# Return the transformer.
		return self


	def transform(self, X=None, load_gm_train=False):
		"""Compute the graph kernel matrix between given and fitted data.

		Parameters
		----------
		X : TYPE
			DESCRIPTION.

		Raises
		------
		ValueError
			DESCRIPTION.

		Returns
		-------
		None.

		"""
		# If `load_gm_train`, load Gram matrix of training data.
		if load_gm_train:
			check_is_fitted(self, '_gm_train')
			self._is_transformed = True
			return self._gm_train  # @todo: copy or not?

		# Check if method "fit" had been called.
		check_is_fitted(self, '_graphs')

		# Validate the input.
		Y = self.validate_input(X)

		# Transform: compute the graph kernel matrix.
		kernel_matrix = self.compute_kernel_matrix(Y)
		self._Y = Y

		# Self transform must appear before the diagonal call on normilization.
		self._is_transformed = True
		if self.normalize:
			X_diag, Y_diag = self.diagonals()
			# Catch FloatingPointError: invalid value encountered in sqrt:
			old_settings = np.seterr(invalid='raise')
			try:
				kernel_matrix /= np.sqrt(np.outer(Y_diag, X_diag))
			except:
				raise
			finally:
				np.seterr(**old_settings)

		return kernel_matrix


	def fit_transform(
			self,
			X,
			save_gm_train: bool = False,
			save_mm_train: bool = False,
	):
		"""Fit and transform: compute Gram matrix on the same data.

		Parameters
		----------
		X : list of graphs
			Input graphs.

		Returns
		-------
		gram_matrix : numpy array, shape = [len(X), len(X)]
			The Gram matrix of X.

		"""
		self.fit(X)

		# Transform: compute Gram matrix.
		gram_matrix = self.compute_kernel_matrix()

		# Normalize.
		if self.normalize:
			self._X_diag = np.diagonal(gram_matrix).copy()
			# Catch FloatingPointError: invalid value encountered in sqrt:
			old_settings = np.seterr(invalid='raise')
			try:
				gram_matrix /= np.sqrt(np.outer(self._X_diag, self._X_diag))
			except:
				# print('Error: invalid value encountered in sqrt.')
				# print('self._X_diag =', self._X_diag)
				raise
			finally:
				np.seterr(**old_settings)

		if save_mm_train or save_gm_train:
			self._gm_train = gram_matrix

		return gram_matrix


	def get_params(
			self,
			with_graphs: bool = False,
			with_ndarray: bool = False,
			check_json_serializable: bool = True
	):
		"""Get parameters for this estimator.

		Parameters
		----------
		with_graphs : bool, optional
			Whether to include the graphs. Default: False.

		with_ndarray : bool, optional
			Whether to include the ndarray. Default: False.

		check_json_serializable : bool, optional
			Whether to check if the parameters are JSON serializable. Default: True.
			todo: maybe this needs to be checked in case some important attributes are
			removed.

		Returns
		-------
		params : dict
			Parameter names mapped to their values.

		Todos
		-----
		It may be better to seperate this method with the __str__ method.
		"""
		# loop over attributes in the object:
		params = dict()
		for key, value in self.__dict__.items():
			cur_params = dict()
			# if the attribute is a list of graphs or a graph:
			if (isinstance(value, list) and len(value) > 0 and \
			    isinstance(value[0], nx.Graph)) or \
					isinstance(value, nx.Graph):
				if with_graphs:
					# add the name(s) and params to dict:
					cur_params[key] = dict()
					cur_params[key]['name'] = value[0].__class__.__name__
					cur_params[key]['params'] = value[0].get_params()
				else:
					continue

			# if the attribute is a numpy array:
			elif isinstance(value, np.ndarray):
				if with_ndarray:
					cur_params[key] = value
				else:
					continue

			# If the attribute is a function:
			elif hasattr(value, '__call__'):
				# If it is a partial function, add its `__str__()`:
				if isinstance(value, functools.partial):
					cur_params[key] = str(value)
				# Otherwise, add its name to dict:
				else:
					cur_params[key] = value.__module__ + '.' + value.__name__

			# If the attribute is a class, add its name and params to dict:
			elif hasattr(value, '__dict__'):
				cur_params[key] = dict()
				cur_params[key]['name'] = value.__class__.__name__
				cur_params[key]['params'] = value.get_params()

			# If the attribute is a basic type, add it to dict:
			elif is_basic_python_type(value, deep=True):
				cur_params[key] = value

			# Otherwise, do nothing.
			else:
				continue

			# todo: SpecialLabel.DUMMY (e.g., COX2 + Path)

			if check_json_serializable:
				# If the current params is serializable, add it to params:
				try:
					import json
					json.dumps(cur_params)
				except TypeError:
					continue

			params[key] = cur_params[key]

		return params


	def set_params(self):
		pass


	def clear_attributes(self):
		if hasattr(self, '_X_diag'):
			delattr(self, '_X_diag')
		if hasattr(self, '_graphs'):
			delattr(self, '_graphs')
		if hasattr(self, '_Y'):
			delattr(self, '_Y')
		if hasattr(self, '_run_time'):
			delattr(self, '_run_time')
		if hasattr(self, '_test_run_time'):
			delattr(self, '_test_run_time')


	def validate_parameters(self):
		"""Validate all parameters for the transformer.

		Returns
		-------
		None.

		"""
		if self.parallel == False:
			self.parallel = None
		elif self.parallel == True:
			self.parallel = 'imap_unordered'
		elif self.parallel is not None and self.parallel != 'imap_unordered':
			raise ValueError('Parallel mode is not set correctly.')

		if self.parallel == 'imap_unordered' and self.n_jobs is None:
			self.n_jobs = multiprocessing.cpu_count()


	def validate_input(self, X):
		"""Validate the given input and raise errors if it is invalid.

		Parameters
		----------
		X : list
			The input to check. Should be a list of graph.

		Raises
		------
		ValueError
			Raise if the input is not correct.

		Returns
		-------
		X : list
			The input. A list of graph.

		"""
		if X is None:
			raise ValueError('Please add graphs before computing.')
		elif not isinstance(X, list):
			raise ValueError('Cannot detect graphs.')
		elif len(X) == 0:
			raise ValueError(
				'The graph list given is empty. No computation will be performed.'
			)

		return X


	def compute_kernel_matrix(self, Y=None):
		"""Compute the kernel matrix between a given target graphs (Y) and
		the fitted graphs (X / self._graphs) or the Gram matrix for the fitted
		graphs (X / self._graphs).

		Parameters
		----------
		Y : list of graphs, optional
			The target graphs. The default is None. If None kernel is computed
			between X and itself.

		Returns
		-------
		kernel_matrix : numpy array, shape = [n_targets, n_inputs]
			The computed kernel matrix.

		"""
		if Y is None:
			# Compute Gram matrix for self._graphs (X).
			kernel_matrix = self._compute_gram_matrix()
		# 			self._gram_matrix_unnorm = np.copy(self._gram_matrix)

		else:
			# Compute kernel matrix between Y and self._graphs (X).
			if self.parallel == 'imap_unordered':
				start_time = time.time()
				kernel_matrix = self._compute_kernel_matrix_imap_unordered(Y)

			elif self.parallel is None:
				Y_copy = ([g.copy() for g in Y] if self.copy_graphs else Y)
				graphs_copy = (
					[g.copy() for g in
					 self._graphs] if self.copy_graphs else self._graphs
				)
				start_time = time.time()
				kernel_matrix = self._compute_kernel_matrix_series(
					Y_copy, graphs_copy
				)

			self._test_run_time = time.time() - start_time
			if self.verbose:
				print(
					'Kernel matrix of size (%d, %d) built in %s seconds.'
					% (len(Y), len(self._graphs), self._test_run_time)
				)

		return kernel_matrix


	def _compute_kernel_matrix_series(self, X, Y):
		"""Compute the kernel matrix between two sets of graphs (X and Y) without parallelization.

		Parameters
		----------
		X, Y : list of graphs
			The input graphs.

		Returns
		-------
		kernel_matrix : numpy array, shape = [n_X, n_Y]
			The computed kernel matrix.

		"""
		kernel_matrix = np.zeros((len(X), len(Y)))

		for i_x, g_x in enumerate(X):
			for i_y, g_y in enumerate(Y):
				kernel_matrix[i_x, i_y] = self.pairwise_kernel(g_x, g_y)

		return kernel_matrix


	def _compute_kernel_matrix_imap_unordered(self, Y):
		"""Compute the kernel matrix between a given target graphs (Y) and
		the fitted graphs (X / self._graphs) using imap unordered parallelization.

		Parameters
		----------
		Y : list of graphs, optional
			The target graphs.

		Returns
		-------
		kernel_matrix : numpy array, shape = [n_targets, n_inputs]
			The computed kernel matrix.

		"""
		raise Exception('Parallelization for kernel matrix is not implemented.')


	def diagonals(self):
		"""Compute the kernel matrix diagonals of the fit/transformed data.

		Returns
		-------
        X_diag : numpy array
            The diagonal of the kernel matrix between the fitted data.
            This consists of each element calculated with itself.

        Y_diag : numpy array
            The diagonal of the kernel matrix, of the transform.
            This consists of each element calculated with itself.

		"""
		# Check if method "fit" had been called.
		check_is_fitted(self, ['_graphs'])

		# Check if the diagonals of X exist.
		try:
			check_is_fitted(self, ['_X_diag'])
		except NotFittedError:
			# Compute diagonals of X.
			self._X_diag = np.empty(shape=(len(self._graphs),))
			graphs = ([g.copy() for g in
			           self._graphs] if self.copy_graphs else self._graphs)
			for i, x in enumerate(graphs):
				self._X_diag[i] = self.pairwise_kernel(x, x)  # @todo: parallel?

		try:
			# If transform has happened, return both diagonals.
			check_is_fitted(self, ['_Y'])
			self._Y_diag = np.empty(shape=(len(self._Y),))
			Y = ([g.copy() for g in self._Y] if self.copy_graphs else self._Y)
			for (i, y) in enumerate(Y):
				self._Y_diag[i] = self.pairwise_kernel(y, y)  # @todo: parallel?

			return self._X_diag, self._Y_diag
		except NotFittedError:
			# Else just return both X_diag
			return self._X_diag


	# 	@abstractmethod
	def pairwise_kernel(self, x, y):
		"""Compute pairwise kernel between two graphs.

		Parameters
		----------
		x, y : NetworkX Graph.
			Graphs bewteen which the kernel is computed.

		Returns
		-------
		kernel: float
			The computed kernel.

# 		Notes
# 		-----
# 		This method is abstract and must be implemented by a subclass.

		"""
		raise NotImplementedError(
			'Pairwise kernel computation is not implemented!'
		)


	##########################################################################
	# The following is the 2nd paradigm to compute kernel matrix. It is
	# simplified and not compatible with `scikit-learn`.
	##########################################################################

	def compute(self, *graphs, **kwargs):
		self.parallel = kwargs.get('parallel', 'imap_unordered')
		self.n_jobs = kwargs.get('n_jobs', multiprocessing.cpu_count())
		self.normalize = kwargs.get('normalize', True)
		self.verbose = kwargs.get('verbose', 2)
		self.copy_graphs = kwargs.get('copy_graphs', True)
		self.save_unnormed = kwargs.get('save_unnormed', True)
		self.validate_parameters()

		# If the inputs is a list of graphs.
		if len(graphs) == 1:
			if not isinstance(graphs[0], list):
				raise Exception('Cannot detect graphs.')
			elif len(graphs[0]) == 0:
				raise Exception(
					'The graph list given is empty. No computation was performed.'
				)
			else:
				if self.copy_graphs:
					self._graphs = [
						g.copy() for g in
						graphs[0]]  # @todo: might be very slow.
				else:
					self._graphs = graphs
				self._gm_train = self._compute_gram_matrix()

				if self.save_unnormed:
					self._gram_matrix_unnorm = np.copy(self._gm_train)
				if self.normalize:
					self._gm_train = normalize_gram_matrix(self._gm_train)
				return self._gm_train, self._run_time

		elif len(graphs) == 2:
			# If the inputs are two graphs.
			if self.is_graph(graphs[0]) and self.is_graph(graphs[1]):
				if self.copy_graphs:
					G0, G1 = graphs[0].copy(), graphs[1].copy()
				else:
					G0, G1 = graphs[0], graphs[1]
				kernel = self._compute_single_kernel(G0, G1)
				return kernel, self._run_time

			# If the inputs are a graph and a list of graphs.
			elif self.is_graph(graphs[0]) and isinstance(graphs[1], list):
				if self.copy_graphs:
					g1 = graphs[0].copy()
					g_list = [g.copy() for g in graphs[1]]
					kernel_list = self._compute_kernel_list(g1, g_list)
				else:
					kernel_list = self._compute_kernel_list(
						graphs[0], graphs[1]
					)
				return kernel_list, self._run_time

			elif isinstance(graphs[0], list) and self.is_graph(graphs[1]):
				if self.copy_graphs:
					g1 = graphs[1].copy()
					g_list = [g.copy() for g in graphs[0]]
					kernel_list = self._compute_kernel_list(g1, g_list)
				else:
					kernel_list = self._compute_kernel_list(
						graphs[1], graphs[0]
					)
				return kernel_list, self._run_time

			else:
				raise Exception('Cannot detect graphs.')

		elif len(graphs) == 0 and self._graphs is None:
			raise Exception('Please add graphs before computing.')

		else:
			raise Exception('Cannot detect graphs.')


	@staticmethod
	def normalize_gm(gram_matrix):
		import warnings
		warnings.warn(
			'gklearn.kernels.graph_kernel.normalize_gm will be deprecated, use '
			'gklearn.utils.normalize_gram_matrix instead',
			DeprecationWarning
		)

		diag = gram_matrix.diagonal().copy()
		for i in range(len(gram_matrix)):
			for j in range(i, len(gram_matrix)):
				gram_matrix[i][j] /= np.sqrt(diag[i] * diag[j])
				gram_matrix[j][i] = gram_matrix[i][j]
		return gram_matrix


	def compute_distance_matrix(self):
		if self._gm_train is None:
			raise Exception(
				'Please compute the Gram matrix before computing distance matrix.'
			)
		dis_mat = np.empty((len(self._gm_train), len(self._gm_train)))
		for i in range(len(self._gm_train)):
			for j in range(i, len(self._gm_train)):
				dis = self._gm_train[i, i] + self._gm_train[j, j] - 2 * \
				      self._gm_train[i, j]
				if dis < 0:
					if dis > -1e-10:
						dis = 0
					else:
						raise ValueError('The distance is negative.')
				dis_mat[i, j] = np.sqrt(dis)
				dis_mat[j, i] = dis_mat[i, j]
		dis_max = np.max(np.max(dis_mat))
		dis_min = np.min(np.min(dis_mat[dis_mat != 0]))
		dis_mean = np.mean(np.mean(dis_mat))
		return dis_mat, dis_max, dis_min, dis_mean


	def _compute_gram_matrix(self):
		if self.parallel == 'imap_unordered':
			start_time = time.time()
			gram_matrix = self._compute_gm_imap_unordered()

		elif self.parallel is None:
			graphs = (
				[g.copy() for g in
				 self._graphs] if self.copy_graphs else self._graphs)

			# todo: this is just a temporary fix for the self loop problem.
			# Remove self loops from the graphs:
			for g in graphs:
				for node in g:
					if g.has_edge(node, node):
						g.remove_edge(node, node)

			start_time = time.time()
			gram_matrix = self._compute_gm_series(graphs)

		else:
			raise Exception('Parallel mode is not set correctly.')

		self._run_time = time.time() - start_time
		if self.verbose:
			print(
				'Gram matrix of size %d built in %s seconds.'
				% (len(self._graphs), self._run_time)
			)

		return gram_matrix


	def _compute_gm_series(self, graphs):
		raise NotImplementedError(
			'The `_compute_gm_series` method needs to be implemented by a subclass.'
		)


	def _compute_gm_imap_unordered(self, graphs):
		raise NotImplementedError(
			'The `_compute_gm_imap_unordered` method needs to be implemented by '
			'a subclass.'
		)


	def _compute_kernel_list(self, g1, g_list):
		start_time = time.time()

		if self.parallel == 'imap_unordered':
			kernel_list = self._compute_kernel_list_imap_unordered(g1, g_list)
		elif self.parallel is None:
			kernel_list = self._compute_kernel_list_series(g1, g_list)
		else:
			raise Exception('Parallel mode is not set correctly.')

		self._run_time = time.time() - start_time
		if self.verbose:
			print(
				'Graph kernel bewteen a graph and a list of %d graphs built in %s seconds.'
				% (len(g_list), self._run_time)
			)

		return kernel_list


	def _compute_kernel_list_series(self, g1, g_list):
		raise NotImplementedError(
			'The `_compute_kernel_list_series` method needs to be implemented by '
			'a subclass.'
		)


	def _compute_kernel_list_imap_unordered(self, g1, g_list):
		raise NotImplementedError(
			'The `_compute_kernel_list_imap_unordered` method needs to be '
			'implemented by a subclass.'
		)


	def _compute_single_kernel(self, g1, g2):
		start_time = time.time()

		kernel = self._compute_single_kernel_series(g1, g2)

		self._run_time = time.time() - start_time
		if self.verbose:
			print(
				'Graph kernel bewteen two graphs built in %s seconds.' % (
					self._run_time)
			)

		return kernel


	def _compute_single_kernel_series(self, g1, g2):
		raise NotImplementedError(
			'The `_compute_single_kernel_series` method needs to be implemented '
			'by a subclass.'
		)


	@staticmethod
	def is_graph(graph):
		if isinstance(graph, nx.Graph):
			return True
		if isinstance(graph, nx.DiGraph):
			return True
		if isinstance(graph, nx.MultiGraph):
			return True
		if isinstance(graph, nx.MultiDiGraph):
			return True
		return False


	@property
	def graphs(self):
		return self._graphs


	# 	@property
	# 	def parallel(self):
	# 		return self.parallel

	# 	@property
	# 	def n_jobs(self):
	# 		return self.n_jobs

	# 	@property
	# 	def verbose(self):
	# 		return self.verbose

	# 	@property
	# 	def normalize(self):
	# 		return self.normalize

	@property
	def run_time(self):
		return self._run_time


	@property
	def test_run_time(self):
		return self._test_run_time


	@property
	def gram_matrix(self):
		return self._gm_train


	@gram_matrix.setter
	def gram_matrix(self, value):
		self._gm_train = value


	@property
	def metric_matrix(self):
		return self._gm_train


	@metric_matrix.setter
	def metric_matrix(self, value):
		self._gm_train = value


	@property
	def gram_matrix_unnorm(self):
		return self._gram_matrix_unnorm


	@gram_matrix_unnorm.setter
	def gram_matrix_unnorm(self, value):
		self._gram_matrix_unnorm = value


	@property
	def n_pairs(self):
		"""
		The number of pairs of graphs between which the kernels are computed.
		"""
		try:
			check_is_fitted(self, '_gm_train')
			return len(self._gm_train) * (len(self._gm_train) + 1) / 2
		except NotFittedError:
			return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar 19 18:17:38 2020

@author: ljia
"""

from enum import Enum, unique


class Options(object):
	"""Contains enums for options employed by ged::GEDEnv.
	"""


	@unique
	class GEDMethod(Enum):
		"""Selects the method.
		"""
# @todo: what is this? #ifdef GUROBI
		F1 = 1                  # Selects ged::F1.
		F2 = 2                 # Selects ged::F2.
		COMPACT_MIP = 3         # Selects ged::CompactMIP.
		BLP_NO_EDGE_LABELS = 4  # Selects ged::BLPNoEdgeLabels.
#endif /* GUROBI */
		BRANCH = 5              # Selects ged::Branch.
		BRANCH_FAST = 6         # Selects ged::BranchFast.
		BRANCH_TIGHT = 7        # Selects ged::BranchTight.
		BRANCH_UNIFORM = 8      # Selects ged::BranchUniform.
		BRANCH_COMPACT = 9      # Selects ged::BranchCompact.
		PARTITION = 10           # Selects ged::Partition.
		HYBRID = 11              # Selects ged::Hybrid.
		RING = 12                # Selects ged::Ring.
		ANCHOR_AWARE_GED = 13    # Selects ged::AnchorAwareGED.
		WALKS = 14               # Selects ged::Walks.
		IPFP = 15                # Selects ged::IPFP
		BIPARTITE = 16           # Selects ged::Bipartite.
		SUBGRAPH = 17            # Selects ged::Subgraph.
		NODE = 18                # Selects ged::Node.
		RING_ML = 19             # Selects ged::RingML.
		BIPARTITE_ML = 20        # Selects ged::BipartiteML.
		REFINE = 21              # Selects ged::Refine.
		BP_BEAM = 22             # Selects ged::BPBeam.
		SIMULATED_ANNEALING = 23 # Selects ged::SimulatedAnnealing.
		HED = 24				 # Selects ged::HED.
		STAR = 25				 # Selects ged::Star.


	@unique
	class EditCosts(Enum):
		"""Selects the edit costs.
		"""
		CHEM_1 = 1      # Selects ged::CHEM1.
		CHEM_2 = 2      # Selects ged::CHEM2.
		CMU = 3         # Selects ged::CMU.
		GREC_1 = 4      # Selects ged::GREC1.
		GREC_2 = 5      # Selects ged::GREC2.
		PROTEIN = 6     # Selects ged::Protein.
		FINGERPRINT = 7 # Selects ged::Fingerprint.
		LETTER = 8      # Selects ged::Letter.
		LETTER2 = 9     # Selects ged:Letter2.
		NON_SYMBOLIC = 10 # Selects ged:NonSymbolic.
		CONSTANT = 11    # Selects ged::Constant.
		
		
	@unique
	class InitType(Enum):
		"""@brief Selects the initialization type of the environment.
		* @details If eager initialization is selected, all edit costs are pre-computed when initializing the environment.
		* Otherwise, they are computed at runtime. If initialization with shuffled copies is selected, shuffled copies of
		* all graphs are created. These copies are used when calling ged::GEDEnv::run_method() with two identical graph IDs.
		* In this case, one of the IDs is internally replaced by the ID of the shuffled copy and the graph is hence
		* compared to an isomorphic but non-identical graph. If initialization without shuffled copies is selected, no shuffled copies
		* are created and calling ged::GEDEnv::run_method() with two identical graph IDs amounts to comparing a graph to itself.
		"""
		LAZY_WITHOUT_SHUFFLED_COPIES = 1  # Lazy initialization, no shuffled graph copies are constructed.
		EAGER_WITHOUT_SHUFFLED_COPIES = 2 # Eager initialization, no shuffled graph copies are constructed.
		LAZY_WITH_SHUFFLED_COPIES = 3     # Lazy initialization, shuffled graph copies are constructed.
		EAGER_WITH_SHUFFLED_COPIES = 4    # Eager initialization, shuffled graph copies are constructed.	
	
	
	@unique
	class AlgorithmState(Enum):
		"""can be used to specify the state of an algorithm.
		"""
		CALLED = 1 # The algorithm has been called.
		INITIALIZED = 2 # The algorithm has been initialized.
		CONVERGED = 3 # The algorithm has converged.
		TERMINATED = 4 # The algorithm has terminated.	


class OptionsStringMap(object):
	
	
	# Map of available computation methods between enum type and string. 
	GEDMethod = {
		"BRANCH": Options.GEDMethod.BRANCH,
		"BRANCH_FAST": Options.GEDMethod.BRANCH_FAST,
		"BRANCH_TIGHT": Options.GEDMethod.BRANCH_TIGHT,
		"BRANCH_UNIFORM": Options.GEDMethod.BRANCH_UNIFORM,
		"BRANCH_COMPACT": Options.GEDMethod.BRANCH_COMPACT,
		"PARTITION": Options.GEDMethod.PARTITION,
		"HYBRID": Options.GEDMethod.HYBRID,
		"RING": Options.GEDMethod.RING,
		"ANCHOR_AWARE_GED": Options.GEDMethod.ANCHOR_AWARE_GED,
		"WALKS": Options.GEDMethod.WALKS,
		"IPFP": Options.GEDMethod.IPFP,
		"BIPARTITE": Options.GEDMethod.BIPARTITE,
		"SUBGRAPH": Options.GEDMethod.SUBGRAPH,
		"NODE": Options.GEDMethod.NODE,
		"RING_ML": Options.GEDMethod.RING_ML,
		"BIPARTITE_ML": Options.GEDMethod.BIPARTITE_ML,
		"REFINE": Options.GEDMethod.REFINE,
		"BP_BEAM": Options.GEDMethod.BP_BEAM,
		"SIMULATED_ANNEALING": Options.GEDMethod.SIMULATED_ANNEALING,
		"HED": Options.GEDMethod.HED,
		"STAR": Options.GEDMethod.STAR,
		# ifdef GUROBI
		"F1": Options.GEDMethod.F1,
		"F2": Options.GEDMethod.F2,
		"COMPACT_MIP": Options.GEDMethod.COMPACT_MIP,
		"BLP_NO_EDGE_LABELS": Options.GEDMethod.BLP_NO_EDGE_LABELS
	}

	
	# Map of available edit cost functions between enum type and string.
	EditCosts = {
		"CHEM_1": Options.EditCosts.CHEM_1,
		"CHEM_2": Options.EditCosts.CHEM_2,
		"CMU": Options.EditCosts.CMU,
		"GREC_1": Options.EditCosts.GREC_1,
		"GREC_2": Options.EditCosts.GREC_2,
		"LETTER": Options.EditCosts.LETTER,
		"LETTER2": Options.EditCosts.LETTER2,
		"NON_SYMBOLIC": Options.EditCosts.NON_SYMBOLIC,
		"FINGERPRINT": Options.EditCosts.FINGERPRINT,
		"PROTEIN": Options.EditCosts.PROTEIN,
		"CONSTANT": Options.EditCosts.CONSTANT
	}
	
	# Map of available initialization types of the environment between enum type and string.
	InitType = {
		"LAZY_WITHOUT_SHUFFLED_COPIES": Options.InitType.LAZY_WITHOUT_SHUFFLED_COPIES,
		"EAGER_WITHOUT_SHUFFLED_COPIES": Options.InitType.EAGER_WITHOUT_SHUFFLED_COPIES,
		"LAZY_WITH_SHUFFLED_COPIES": Options.InitType.LAZY_WITH_SHUFFLED_COPIES,
		"LAZY_WITH_SHUFFLED_COPIES": Options.InitType.LAZY_WITH_SHUFFLED_COPIES
	}
	

@unique
class AlgorithmState(Enum):
	"""can be used to specify the state of an algorithm.
	"""
	CALLED = 1 # The algorithm has been called.
	INITIALIZED = 2 # The algorithm has been initialized.
	CONVERGED = 3 # The algorithm has converged.
	TERMINATED = 4 # The algorithm has terminated.
	



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wen May 27 14:27:15 2020

@author: ljia
"""
import numpy as np
import csv
import os
import os.path
from gklearn.utils import Dataset
from gklearn.preimage import MedianPreimageGenerator
from gklearn.utils import normalize_gram_matrix
from gklearn.utils import split_dataset_by_target
from gklearn.preimage.utils import compute_k_dis
from gklearn.utils.graphfiles import saveGXL
import networkx as nx
	

def remove_best_graph(ds_name, mpg_options, kernel_options, ged_options, mge_options, save_results=True, save_medians=True, plot_medians=True, load_gm='auto', dir_save='', irrelevant_labels=None, edge_required=False, cut_range=None):
	"""Remove the best graph from the median set w.r.t. distance in kernel space, and to see if it is possible to generate the removed graph using the graphs left in the median set.
	"""
	# 1. get dataset.
	print('1. getting dataset...')
	dataset_all = Dataset()
	dataset_all.load_predefined_dataset(ds_name)
	dataset_all.trim_dataset(edge_required=edge_required)
	if irrelevant_labels is not None:
		dataset_all.remove_labels(**irrelevant_labels)
	if cut_range is not None:
		dataset_all.cut_graphs(cut_range)
	datasets = split_dataset_by_target(dataset_all)

	if save_results:
		# create result files.
		print('creating output files...')
		fn_output_detail, fn_output_summary = _init_output_file(ds_name, kernel_options['name'], mpg_options['fit_method'], dir_save)
	else:
		fn_output_detail, fn_output_summary = None, None
		
	# 2. compute/load Gram matrix a priori.
	print('2. computing/loading Gram matrix...')
	gram_matrix_unnorm_list, time_precompute_gm_list = _get_gram_matrix(load_gm, dir_save, ds_name, kernel_options, datasets)
		
	sod_sm_list = []
	sod_gm_list = []
	dis_k_sm_list = []
	dis_k_gm_list = []
	dis_k_gi_min_list = []
	time_optimize_ec_list = []
	time_generate_list = []
	time_total_list = []
	itrs_list = []
	converged_list = []
	num_updates_ecc_list = []
	mge_decrease_order_list = []
	mge_increase_order_list = []
	mge_converged_order_list = []
	nb_sod_sm2gm = [0, 0, 0]
	nb_dis_k_sm2gm = [0, 0, 0]
	nb_dis_k_gi2sm = [0, 0, 0]
	nb_dis_k_gi2gm = [0, 0, 0]
	dis_k_max_list = []	
	dis_k_min_list = []
	dis_k_mean_list = []
	best_dis_list = []
	print('starting experiment for each class of target...')
	idx_offset = 0
	for idx, dataset in enumerate(datasets):
		target = dataset.targets[0]
		print('\ntarget =', target, '\n')
#		if target != 1:
# 			continue
		
		num_graphs = len(dataset.graphs)
		if num_graphs < 2:
			print('\nnumber of graphs = ', num_graphs, ', skip.\n')
			idx_offset += 1
			continue

		# 3. get the best graph and remove it from median set.
		print('3. getting and removing the best graph...')
		gram_matrix_unnorm = gram_matrix_unnorm_list[idx - idx_offset]
		best_index, best_dis, best_graph = _get_best_graph([g.copy() for g in dataset.graphs], normalize_gram_matrix(gram_matrix_unnorm.copy()))
		median_set_new = [dataset.graphs[i] for i in range(len(dataset.graphs)) if i != best_index]
		num_graphs -= 1
		if num_graphs == 1:
			continue		
		best_dis_list.append(best_dis)
		
		dataset.load_graphs(median_set_new, targets=None)
		gram_matrix_unnorm_new = np.delete(gram_matrix_unnorm, best_index, axis=0)
		gram_matrix_unnorm_new = np.delete(gram_matrix_unnorm_new, best_index, axis=1)
			
		# 4. set parameters.
		print('4. initializing mpg and setting parameters...')	
		mpg_options['gram_matrix_unnorm'] = gram_matrix_unnorm_new
		mpg_options['runtime_precompute_gm'] = time_precompute_gm_list[idx - idx_offset]
		mpg = MedianPreimageGenerator()
		mpg.dataset = dataset
		mpg.set_options(**mpg_options.copy())
		mpg.kernel_options = kernel_options.copy()
		mpg.ged_options = ged_options.copy()
		mpg.mge_options = mge_options.copy()

		# 5. compute median preimage.
		print('5. computing median preimage...')
		mpg.run()
		results = mpg.get_results()
		
		# 6. compute pairwise kernel distances.
		print('6. computing pairwise kernel distances...')
		_, dis_k_max, dis_k_min, dis_k_mean = mpg.graph_kernel.compute_distance_matrix()
		dis_k_max_list.append(dis_k_max)
		dis_k_min_list.append(dis_k_min)
		dis_k_mean_list.append(dis_k_mean)
		
		# 7. save results (and median graphs).
		print('7. saving results (and median graphs)...')
		# write result detail.
		if save_results:
			print('writing results to files...')
			sod_sm2gm = get_relations(np.sign(results['sod_gen_median'] - results['sod_set_median']))
			dis_k_sm2gm = get_relations(np.sign(results['k_dis_gen_median'] - results['k_dis_set_median']))
			dis_k_gi2sm = get_relations(np.sign(results['k_dis_set_median'] - results['k_dis_dataset']))
			dis_k_gi2gm = get_relations(np.sign(results['k_dis_gen_median'] - results['k_dis_dataset']))

			f_detail = open(dir_save + fn_output_detail, 'a')
			csv.writer(f_detail).writerow([ds_name, kernel_options['name'], 
					  ged_options['edit_cost'], ged_options['method'], 
					  ged_options['attr_distance'], mpg_options['fit_method'], 
					  num_graphs, target, 1, 
					  results['sod_set_median'], results['sod_gen_median'],
					  results['k_dis_set_median'], results['k_dis_gen_median'], 
					  results['k_dis_dataset'], best_dis, best_index,
					  sod_sm2gm, dis_k_sm2gm, 
					  dis_k_gi2sm, dis_k_gi2gm,	results['edit_cost_constants'],
					  results['runtime_precompute_gm'], results['runtime_optimize_ec'], 
					  results['runtime_generate_preimage'], results['runtime_total'],
					  results['itrs'], results['converged'],
					  results['num_updates_ecc'],
					  results['mge']['num_decrease_order'] > 0, # @todo: not suitable for multi-start mge
					  results['mge']['num_increase_order'] > 0,
					  results['mge']['num_converged_descents'] > 0])
			f_detail.close()
		
			# compute result summary.
			sod_sm_list.append(results['sod_set_median'])
			sod_gm_list.append(results['sod_gen_median'])
			dis_k_sm_list.append(results['k_dis_set_median'])
			dis_k_gm_list.append(results['k_dis_gen_median'])
			dis_k_gi_min_list.append(results['k_dis_dataset'])
			time_precompute_gm_list.append(results['runtime_precompute_gm'])
			time_optimize_ec_list.append(results['runtime_optimize_ec'])
			time_generate_list.append(results['runtime_generate_preimage'])
			time_total_list.append(results['runtime_total'])
			itrs_list.append(results['itrs'])
			converged_list.append(results['converged'])
			num_updates_ecc_list.append(results['num_updates_ecc'])
			mge_decrease_order_list.append(results['mge']['num_decrease_order'] > 0)
			mge_increase_order_list.append(results['mge']['num_increase_order'] > 0)
			mge_converged_order_list.append(results['mge']['num_converged_descents'] > 0)
			# # SOD SM -> GM
			if results['sod_set_median'] > results['sod_gen_median']:
				nb_sod_sm2gm[0] += 1
	#			repeats_better_sod_sm2gm.append(1)
			elif results['sod_set_median'] == results['sod_gen_median']:
				nb_sod_sm2gm[1] += 1
			elif results['sod_set_median'] < results['sod_gen_median']:
				nb_sod_sm2gm[2] += 1
			# # dis_k SM -> GM
			if results['k_dis_set_median'] > results['k_dis_gen_median']:
				nb_dis_k_sm2gm[0] += 1
	#			repeats_better_dis_k_sm2gm.append(1)
			elif results['k_dis_set_median'] == results['k_dis_gen_median']:
				nb_dis_k_sm2gm[1] += 1
			elif results['k_dis_set_median'] < results['k_dis_gen_median']:
				nb_dis_k_sm2gm[2] += 1
			# # dis_k gi -> SM
			if results['k_dis_dataset'] > results['k_dis_set_median']:
				nb_dis_k_gi2sm[0] += 1
	#			repeats_better_dis_k_gi2sm.append(1)
			elif results['k_dis_dataset'] == results['k_dis_set_median']:
				nb_dis_k_gi2sm[1] += 1
			elif results['k_dis_dataset'] < results['k_dis_set_median']:
				nb_dis_k_gi2sm[2] += 1
			# # dis_k gi -> GM
			if results['k_dis_dataset'] > results['k_dis_gen_median']:
				nb_dis_k_gi2gm[0] += 1
	#			repeats_better_dis_k_gi2gm.append(1)
			elif results['k_dis_dataset'] == results['k_dis_gen_median']:
				nb_dis_k_gi2gm[1] += 1
			elif results['k_dis_dataset'] < results['k_dis_gen_median']:
				nb_dis_k_gi2gm[2] += 1

			# write result summary for each letter.
			f_summary = open(dir_save + fn_output_summary, 'a')
			csv.writer(f_summary).writerow([ds_name, kernel_options['name'], 
					  ged_options['edit_cost'], ged_options['method'], 
					  ged_options['attr_distance'], mpg_options['fit_method'], 
					  num_graphs, target,
					  results['sod_set_median'], results['sod_gen_median'],
					  results['k_dis_set_median'], results['k_dis_gen_median'], 
					  results['k_dis_dataset'], best_dis, best_index,
					  sod_sm2gm, dis_k_sm2gm, 
					  dis_k_gi2sm, dis_k_gi2gm, 
					  results['runtime_precompute_gm'], results['runtime_optimize_ec'], 
					  results['runtime_generate_preimage'], results['runtime_total'],
					  results['itrs'], results['converged'],
					  results['num_updates_ecc'], 
					  results['mge']['num_decrease_order'] > 0, # @todo: not suitable for multi-start mge
					  results['mge']['num_increase_order'] > 0,
					  results['mge']['num_converged_descents'] > 0, 
					  nb_sod_sm2gm, 
					  nb_dis_k_sm2gm, nb_dis_k_gi2sm, nb_dis_k_gi2gm])
			f_summary.close()
			 
		# save median graphs.
		if save_medians:
			os.makedirs(dir_save + 'medians/', exist_ok=True)
			print('Saving median graphs to files...')
			fn_pre_sm = dir_save + 'medians/set_median.' + mpg_options['fit_method'] + '.nbg' + str(num_graphs) + '.y' + str(target) + '.repeat' + str(1)
			saveGXL(mpg.set_median, fn_pre_sm + '.gxl', method='default', 
					node_labels=dataset.node_labels, edge_labels=dataset.edge_labels,	
					node_attrs=dataset.node_attrs, edge_attrs=dataset.edge_attrs)
			fn_pre_gm = dir_save + 'medians/gen_median.' + mpg_options['fit_method'] + '.nbg' + str(num_graphs) + '.y' + str(target) + '.repeat' + str(1)
			saveGXL(mpg.gen_median, fn_pre_gm + '.gxl', method='default',
		  			node_labels=dataset.node_labels, edge_labels=dataset.edge_labels,	
					node_attrs=dataset.node_attrs, edge_attrs=dataset.edge_attrs)
			fn_best_dataset = dir_save + 'medians/g_best_dataset.' + mpg_options['fit_method'] + '.nbg' + str(num_graphs) + '.y' + str(target) + '.repeat' + str(1)
			saveGXL(best_graph, fn_best_dataset + '.gxl', method='default',
		  			node_labels=dataset.node_labels, edge_labels=dataset.edge_labels,	
					node_attrs=dataset.node_attrs, edge_attrs=dataset.edge_attrs)
			fn_best_median_set = dir_save + 'medians/g_best_median_set.' + mpg_options['fit_method'] + '.nbg' + str(num_graphs) + '.y' + str(target) + '.repeat' + str(1)
			saveGXL(mpg.best_from_dataset, fn_best_median_set + '.gxl', method='default',
		  			node_labels=dataset.node_labels, edge_labels=dataset.edge_labels,	
					node_attrs=dataset.node_attrs, edge_attrs=dataset.edge_attrs)
		
		# plot median graphs.
		if plot_medians and save_medians:
			if ged_options['edit_cost'] == 'LETTER2' or ged_options['edit_cost'] == 'LETTER' or ds_name == 'Letter-high' or ds_name == 'Letter-med' or ds_name == 'Letter-low':			
				draw_Letter_graph(mpg.set_median, fn_pre_sm)
				draw_Letter_graph(mpg.gen_median, fn_pre_gm)
				draw_Letter_graph(mpg.best_from_dataset, fn_best_dataset)

	# write result summary for each letter. 
	if save_results:
		sod_sm_mean = np.mean(sod_sm_list)
		sod_gm_mean = np.mean(sod_gm_list)
		dis_k_sm_mean = np.mean(dis_k_sm_list)
		dis_k_gm_mean = np.mean(dis_k_gm_list)
		dis_k_gi_min_mean = np.mean(dis_k_gi_min_list)
		best_dis_mean = np.mean(best_dis_list)
		time_precompute_gm_mean = np.mean(time_precompute_gm_list)
		time_optimize_ec_mean = np.mean(time_optimize_ec_list)
		time_generate_mean = np.mean(time_generate_list)
		time_total_mean = np.mean(time_total_list)
		itrs_mean = np.mean(itrs_list)
		num_converged = np.sum(converged_list)
		num_updates_ecc_mean = np.mean(num_updates_ecc_list)
		num_mge_decrease_order = np.sum(mge_decrease_order_list)
		num_mge_increase_order = np.sum(mge_increase_order_list)
		num_mge_converged = np.sum(mge_converged_order_list)
		sod_sm2gm_mean = get_relations(np.sign(sod_gm_mean - sod_sm_mean))
		dis_k_sm2gm_mean = get_relations(np.sign(dis_k_gm_mean - dis_k_sm_mean))
		dis_k_gi2sm_mean = get_relations(np.sign(dis_k_sm_mean - dis_k_gi_min_mean))
		dis_k_gi2gm_mean = get_relations(np.sign(dis_k_gm_mean - dis_k_gi_min_mean))
		f_summary = open(dir_save + fn_output_summary, 'a')
		csv.writer(f_summary).writerow([ds_name, kernel_options['name'], 
				  ged_options['edit_cost'], ged_options['method'], 
				  ged_options['attr_distance'], mpg_options['fit_method'], 
				  num_graphs, 'all',
				  sod_sm_mean, sod_gm_mean, dis_k_sm_mean, dis_k_gm_mean,
				  dis_k_gi_min_mean, best_dis_mean, '-', 
				  sod_sm2gm_mean, dis_k_sm2gm_mean, 
				  dis_k_gi2sm_mean, dis_k_gi2gm_mean,
				  time_precompute_gm_mean, time_optimize_ec_mean,
				  time_generate_mean, time_total_mean, itrs_mean, 
				  num_converged, num_updates_ecc_mean,
				  num_mge_decrease_order, num_mge_increase_order,
				  num_mge_converged])
		f_summary.close()
		
	# save total pairwise kernel distances.
	dis_k_max = np.max(dis_k_max_list)
	dis_k_min = np.min(dis_k_min_list)
	dis_k_mean = np.mean(dis_k_mean_list)
	print('The maximum pairwise distance in kernel space:', dis_k_max)
	print('The minimum pairwise distance in kernel space:', dis_k_min)
	print('The average pairwise distance in kernel space:', dis_k_mean)

	print('\ncomplete.\n')	


def _get_best_graph(Gn, gram_matrix):
	k_dis_list = []
	for idx in range(len(Gn)):
		k_dis_list.append(compute_k_dis(idx, range(0, len(Gn)), [1 / len(Gn)] * len(Gn), gram_matrix, withterm3=False))
	best_index = np.argmin(k_dis_list)
	best_dis = k_dis_list[best_index]
	best_graph = Gn[best_index].copy()
	return best_index, best_dis, best_graph 


def get_relations(sign):
	if sign == -1:
		return 'better'
	elif sign == 0:
		return 'same'
	elif sign == 1:
		return 'worse'


def _get_gram_matrix(load_gm, dir_save, ds_name, kernel_options, datasets):
	if load_gm == 'auto':
		gm_fname = dir_save + 'gram_matrix_unnorm.' + ds_name + '.' + kernel_options['name'] + '.gm.npz'
		gmfile_exist = os.path.isfile(os.path.abspath(gm_fname))
		if gmfile_exist:
			gmfile = np.load(gm_fname, allow_pickle=True) # @todo: may not be safe.
			gram_matrix_unnorm_list = [item for item in gmfile['gram_matrix_unnorm_list']]
			time_precompute_gm_list = gmfile['run_time_list'].tolist()
		else:
			gram_matrix_unnorm_list = []
			time_precompute_gm_list = []
			for dataset in datasets:
				gram_matrix_unnorm, time_precompute_gm = _compute_gram_matrix_unnorm(dataset, kernel_options)
				gram_matrix_unnorm_list.append(gram_matrix_unnorm)
				time_precompute_gm_list.append(time_precompute_gm)
			np.savez(dir_save + 'gram_matrix_unnorm.' + ds_name + '.' + kernel_options['name'] + '.gm', gram_matrix_unnorm_list=gram_matrix_unnorm_list, run_time_list=time_precompute_gm_list)
	elif not load_gm:
		gram_matrix_unnorm_list = []
		time_precompute_gm_list = []
		for dataset in datasets:
			gram_matrix_unnorm, time_precompute_gm = _compute_gram_matrix_unnorm(dataset, kernel_options)
			gram_matrix_unnorm_list.append(gram_matrix_unnorm)
			time_precompute_gm_list.append(time_precompute_gm)
		np.savez(dir_save + 'gram_matrix_unnorm.' + ds_name + '.' + kernel_options['name'] + '.gm', gram_matrix_unnorm_list=gram_matrix_unnorm_list, run_time_list=time_precompute_gm_list)
	else:
		gm_fname = dir_save + 'gram_matrix_unnorm.' + ds_name + '.' + kernel_options['name'] + '.gm.npz'
		gmfile = np.load(gm_fname, allow_pickle=True) # @todo: may not be safe.
		gram_matrix_unnorm_list = [item for item in gmfile['gram_matrix_unnorm_list']]
		time_precompute_gm_list = gmfile['run_time_list'].tolist()
		
	return gram_matrix_unnorm_list, time_precompute_gm_list


def _get_graph_kernel(dataset, kernel_options):
	from gklearn.utils.utils import get_graph_kernel_by_name
	graph_kernel = get_graph_kernel_by_name(kernel_options['name'], 
				  node_labels=dataset.node_labels,
				  edge_labels=dataset.edge_labels, 
				  node_attrs=dataset.node_attrs,
				  edge_attrs=dataset.edge_attrs,
				  ds_infos=dataset.get_dataset_infos(keys=['directed']),
				  kernel_options=kernel_options)
	return graph_kernel
		
		
def _compute_gram_matrix_unnorm(dataset, kernel_options):
	from gklearn.utils.utils import get_graph_kernel_by_name
	graph_kernel = get_graph_kernel_by_name(kernel_options['name'], 
				  node_labels=dataset.node_labels,
				  edge_labels=dataset.edge_labels, 
				  node_attrs=dataset.node_attrs,
				  edge_attrs=dataset.edge_attrs,
				  ds_infos=dataset.get_dataset_infos(keys=['directed']),
				  kernel_options=kernel_options)
	
	gram_matrix, run_time = graph_kernel.compute(dataset.graphs, **kernel_options)
	gram_matrix_unnorm = graph_kernel.gram_matrix_unnorm
		
	return gram_matrix_unnorm, run_time
		
		
def _init_output_file(ds_name, gkernel, fit_method, dir_output):
	os.makedirs(dir_output, exist_ok=True)
	fn_output_detail = 'results_detail.' + ds_name + '.' + gkernel + '.csv'
	f_detail = open(dir_output + fn_output_detail, 'a')
	csv.writer(f_detail).writerow(['dataset', 'graph kernel', 'edit cost', 
			  'GED method', 'attr distance', 'fit method', 'num graphs', 
			  'target', 'repeat', 'SOD SM', 'SOD GM', 'dis_k SM', 'dis_k GM',
			  'min dis_k gi', 'best kernel dis', 'best graph index', 
			  'SOD SM -> GM', 'dis_k SM -> GM', 'dis_k gi -> SM', 
			  'dis_k gi -> GM', 'edit cost constants', 'time precompute gm',
			  'time optimize ec', 'time generate preimage', 'time total',
			  'itrs', 'converged', 'num updates ecc', 'mge decrease order', 
			  'mge increase order', 'mge converged'])
	f_detail.close()
	
	fn_output_summary = 'results_summary.' + ds_name + '.' + gkernel + '.csv'
	f_summary = open(dir_output + fn_output_summary, 'a')
	csv.writer(f_summary).writerow(['dataset', 'graph kernel', 'edit cost', 
			  'GED method', 'attr distance', 'fit method', 'num graphs', 
			  'target', 'SOD SM', 'SOD GM', 'dis_k SM', 'dis_k GM',
			  'min dis_k gi', 'best kernel dis', 'best graph index',
			  'SOD SM -> GM', 'dis_k SM -> GM', 'dis_k gi -> SM', 
			  'dis_k gi -> GM', 'time precompute gm', 'time optimize ec', 
			  'time generate preimage', 'time total', 'itrs', 'num converged', 
			  'num updates ecc', 'mge num decrease order', 'mge num increase order', 
			  'mge num converged', '# SOD SM -> GM', '# dis_k SM -> GM', 
			  '# dis_k gi -> SM', '# dis_k gi -> GM'])
	f_summary.close()
	
	return fn_output_detail, fn_output_summary


#Dessin median courrant
def draw_Letter_graph(graph, file_prefix):
	import matplotlib
	matplotlib.use('agg')
	import matplotlib.pyplot as plt
	plt.figure()
	pos = {}
	for n in graph.nodes:
		pos[n] = np.array([float(graph.nodes[n]['x']),float(graph.nodes[n]['y'])])
	nx.draw_networkx(graph, pos)
	plt.savefig(file_prefix + '.eps', format='eps', dpi=300)
#	plt.show()
	plt.clf()
	plt.close()


# -*-coding:utf-8 -*-
"""
model learning.
"""

# info
__version__ = "0.2"
__author__  = "Linlin Jia"
__date__    = "November 2020"


from gklearn.model_learning.nested_cv import NestedCV
from gklearn.model_learning.workflow import Workflow
from gklearn.model_learning.parameters import dichotomous_permutation